@@ -1,50 +1,64 @@
 /*******************************************************************************
   Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  
-  (c) Copyright 1996 - 2003 Gary Henderson (gary.henderson@ntlworld.com) and
+  (c) Copyright 1996 - 2002 Gary Henderson (gary.henderson@ntlworld.com) and
                             Jerremy Koot (jkoot@snes9x.com)
 
-  (c) Copyright 2002 - 2003 Matthew Kendora and
-                            Brad Jorsch (anomie@users.sourceforge.net)
- 
+  (c) Copyright 2001 - 2004 John Weidman (jweidman@slip.net)
+
+  (c) Copyright 2002 - 2004 Brad Jorsch (anomie@users.sourceforge.net),
+                            funkyass (funkyass@spam.shaw.ca),
+                            Joel Yliluoma (http://iki.fi/bisqwit/)
+                            Kris Bleakley (codeviolation@hotmail.com),
+                            Matthew Kendora,
+                            Nach (n-a-c-h@users.sourceforge.net),
+                            Peter Bortas (peter@bortas.org) and
+                            zones (kasumitokoduck@yahoo.com)
 
-                      
   C4 x86 assembler and some C emulation code
   (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
-                            _Demo_ (_demo_@zsnes.com), and
-                            Nach (n-a-c-h@users.sourceforge.net)
-                                          
+                            _Demo_ (_demo_@zsnes.com), and Nach
+
   C4 C++ code
   (c) Copyright 2003 Brad Jorsch
 
   DSP-1 emulator code
-  (c) Copyright 1998 - 2003 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
-                            John Weidman (jweidman@slip.net),
-                            neviksti (neviksti@hotmail.com), and
-                            Kris Bleakley (stinkfish@bigpond.com)
- 
+  (c) Copyright 1998 - 2004 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
+                            John Weidman, neviksti (neviksti@hotmail.com),
+                            Kris Bleakley, Andreas Naive
+
   DSP-2 emulator code
   (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                      Lord Nightmare (lord_nightmare@users.sourceforge.net
 
   OBC1 emulator code
-  (c) Copyright 2001 - 2003 zsKnight, pagefault (pagefault@zsnes.com)
+  (c) Copyright 2001 - 2004 zsKnight, pagefault (pagefault@zsnes.com) and
+                            Kris Bleakley
   Ported from x86 assembler to C by sanmaiwashi
 
   SPC7110 and RTC C++ emulator code
   (c) Copyright 2002 Matthew Kendora with research by
                      zsKnight, John Weidman, and Dark Force
 
+  S-DD1 C emulator code
+  (c) Copyright 2003 Brad Jorsch with research by
+                     Andreas Naive and John Weidman
+ 
   S-RTC C emulator code
   (c) Copyright 2001 John Weidman
   
+  ST010 C++ emulator code
+  (c) Copyright 2003 Feather, Kris Bleakley, John Weidman and Matthew Kendora
+
   Super FX x86 assembler emulator code 
   (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 
 
   Super FX C emulator code 
-  (c) Copyright 1997 - 1999 Ivar and Gary Henderson.
+  (c) Copyright 1997 - 1999 Ivar, Gary Henderson and John Weidman
 
 
+  SH assembler code partly based on x86 assembler code
+  (c) Copyright 2002 - 2004 Marcus Comstedt (marcus@mc.pp.se) 
 
  
   Specific ports contains the works of other authors. See headers in
@@ -110,6 +124,10 @@ bool8 S9xInitAPU ()
 		S9xDeinitAPU ();
 		return (FALSE);
     }
+
+	memset(IAPU.RAM, 0, 0x10000);
+	memset(IAPU.ShadowRAM, 0, 0x10000);
+	memset(IAPU.CachedSamples, 0, 0x40000);
 	
     return (TRUE);
 }
@@ -175,6 +193,8 @@ void S9xResetAPU ()
     IAPU.WaitAddress2 = NULL;
     IAPU.WaitCounter = 0;
 #endif
+	IAPU.NextAPUTimerPos = 0;
+	IAPU.APUTimerCounter = 0;
     APU.ShowROM = TRUE;
     IAPU.RAM [0xf1] = 0x80;
 		
@@ -895,6 +915,64 @@ void S9xSetAPUTimer (uint16 Address, uint8 byte)
     }
 }
 
+void S9xUpdateAPUTimer (void)
+{
+	while (CPU.Cycles * 10000L >= IAPU.NextAPUTimerPos)
+	//if (CPU.Cycles * 10000L >= IAPU.NextAPUTimerPos)
+	{
+		//APU_EXECUTE();
+		
+		IAPU.NextAPUTimerPos += SNES_APUTIMER2_CYCLEx10000;		
+		
+		if (APU.TimerEnabled [2])
+		{
+			APU.Timer [2] ++;
+			if (APU.Timer [2] >= APU.TimerTarget [2])
+			{
+			    IAPU.RAM [0xff] = (IAPU.RAM [0xff] + 1) & 0xf;
+			    APU.Timer [2] = 0;
+			#ifdef SPC700_SHUTDOWN		
+			    IAPU.WaitCounter++;
+			    IAPU.APUExecuting = TRUE;
+			#endif		
+			}
+		}
+
+		if (++IAPU.APUTimerCounter == 8)
+		{
+			IAPU.APUTimerCounter = 0;
+			
+			if (APU.TimerEnabled [0])
+			{
+			    APU.Timer [0]++;
+			    if (APU.Timer [0] >= APU.TimerTarget [0])
+			    {
+					IAPU.RAM [0xfd] = (IAPU.RAM [0xfd] + 1) & 0xf;
+					APU.Timer [0] = 0;
+				#ifdef SPC700_SHUTDOWN		
+					IAPU.WaitCounter++;
+					IAPU.APUExecuting = TRUE;
+				#endif		    
+			    }
+			}
+
+			if (APU.TimerEnabled [1])
+			{
+			    APU.Timer [1]++;
+			    if (APU.Timer [1] >= APU.TimerTarget [1])
+			    {
+					IAPU.RAM [0xfe] = (IAPU.RAM [0xfe] + 1) & 0xf;
+					APU.Timer [1] = 0;
+				#ifdef SPC700_SHUTDOWN		
+					IAPU.WaitCounter++;
+					IAPU.APUExecuting = TRUE;
+				#endif		    
+			    }
+			}
+		}
+	}
+}
+
 uint8 S9xGetAPUDSP ()
 {
     uint8 reg = IAPU.RAM [0xf2] & 0x7f;
@@ -938,3 +1016,4 @@ uint8 S9xGetAPUDSP ()
     }
     return (byte);
 }
+
