@@ -1,50 +1,64 @@
 /*******************************************************************************
   Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  
-  (c) Copyright 1996 - 2003 Gary Henderson (gary.henderson@ntlworld.com) and
+  (c) Copyright 1996 - 2002 Gary Henderson (gary.henderson@ntlworld.com) and
                             Jerremy Koot (jkoot@snes9x.com)
 
-  (c) Copyright 2002 - 2003 Matthew Kendora and
-                            Brad Jorsch (anomie@users.sourceforge.net)
- 
+  (c) Copyright 2001 - 2004 John Weidman (jweidman@slip.net)
+
+  (c) Copyright 2002 - 2004 Brad Jorsch (anomie@users.sourceforge.net),
+                            funkyass (funkyass@spam.shaw.ca),
+                            Joel Yliluoma (http://iki.fi/bisqwit/)
+                            Kris Bleakley (codeviolation@hotmail.com),
+                            Matthew Kendora,
+                            Nach (n-a-c-h@users.sourceforge.net),
+                            Peter Bortas (peter@bortas.org) and
+                            zones (kasumitokoduck@yahoo.com)
 
-                      
   C4 x86 assembler and some C emulation code
   (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
-                            _Demo_ (_demo_@zsnes.com), and
-                            Nach (n-a-c-h@users.sourceforge.net)
-                                          
+                            _Demo_ (_demo_@zsnes.com), and Nach
+
   C4 C++ code
   (c) Copyright 2003 Brad Jorsch
 
   DSP-1 emulator code
-  (c) Copyright 1998 - 2003 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
-                            John Weidman (jweidman@slip.net),
-                            neviksti (neviksti@hotmail.com), and
-                            Kris Bleakley (stinkfish@bigpond.com)
- 
+  (c) Copyright 1998 - 2004 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
+                            John Weidman, neviksti (neviksti@hotmail.com),
+                            Kris Bleakley, Andreas Naive
+
   DSP-2 emulator code
   (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                      Lord Nightmare (lord_nightmare@users.sourceforge.net
 
   OBC1 emulator code
-  (c) Copyright 2001 - 2003 zsKnight, pagefault (pagefault@zsnes.com)
+  (c) Copyright 2001 - 2004 zsKnight, pagefault (pagefault@zsnes.com) and
+                            Kris Bleakley
   Ported from x86 assembler to C by sanmaiwashi
 
   SPC7110 and RTC C++ emulator code
   (c) Copyright 2002 Matthew Kendora with research by
                      zsKnight, John Weidman, and Dark Force
 
+  S-DD1 C emulator code
+  (c) Copyright 2003 Brad Jorsch with research by
+                     Andreas Naive and John Weidman
+ 
   S-RTC C emulator code
   (c) Copyright 2001 John Weidman
   
+  ST010 C++ emulator code
+  (c) Copyright 2003 Feather, Kris Bleakley, John Weidman and Matthew Kendora
+
   Super FX x86 assembler emulator code 
   (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 
 
   Super FX C emulator code 
-  (c) Copyright 1997 - 1999 Ivar and Gary Henderson.
+  (c) Copyright 1997 - 1999 Ivar, Gary Henderson and John Weidman
 
 
+  SH assembler code partly based on x86 assembler code
+  (c) Copyright 2002 - 2004 Marcus Comstedt (marcus@mc.pp.se) 
 
  
   Specific ports contains the works of other authors. See headers in
@@ -85,6 +99,7 @@
 #include "apu.h"
 #include "sa1.h"
 #include "spc7110.h"
+#include "dsp1.h"
 
 START_EXTERN_C
 extern uint8 A1, A2, A3, A4, W1, W2, W3, W4;
@@ -2555,6 +2570,7 @@ inline void CPUShutdown()
 	    if (Settings.SA1)
 		S9xSA1ExecuteDuringSleep ();
 	    CPU.Cycles = CPU.NextEvent;
+		S9xUpdateAPUTimer();
 	    if (IAPU.APUExecuting)
 	    {
 		ICPU.CPUExecuting = FALSE;
@@ -2890,29 +2906,76 @@ static void OpEA (void)
 #define PushB(b)\
     S9xSetByte (b, Registers.S.W--);
 
+#define PushBE(b)\
+	S9xSetByte (b, Registers.S.W--);\
+	Registers.SH=0x01;
+
+
 #define PushW(w) \
     S9xSetByte ((w)>>8, Registers.S.W);\
 	S9xSetByte ((w)&0xff, (Registers.S.W - 1)&0xFFFF);\
     Registers.S.W -= 2;
 
+#define PushWE(w) \
+    S9xSetByte ((w)>>8, Registers.S.W--);\
+	Registers.SH=0x01;\
+	S9xSetByte ((w)&0xff, (Registers.S.W--)&0xFFFF);\
+    Registers.SH = 0x01;
+
+#define PushWENew(w) \
+    S9xSetByte ((w)>>8, Registers.S.W--);\
+	S9xSetByte ((w)&0xff, (Registers.S.W--)&0xFFFF);\
+    Registers.SH = 0x01;
+
+//PEA NL
+static void OpF4E1 (void)
+{
+    Absolute (NONE);
+    PushWENew ((unsigned short)OpAddress);
+}
+
 static void OpF4 (void)
 {
     Absolute (NONE);
     PushW ((unsigned short)OpAddress);
 }
 
+//PEI NL
+static void OpD4E1 (void)
+{
+    DirectIndirect (NONE);
+    PushWENew ((unsigned short)OpAddress);
+}
+
 static void OpD4 (void)
 {
     DirectIndirect (NONE);
     PushW ((unsigned short)OpAddress);
 }
 
+//PER NL
+static void Op62E1 (void)
+{
+    RelativeLong (NONE);
+    PushWENew ((unsigned short)OpAddress);
+}
+
 static void Op62 (void)
 {
     RelativeLong (NONE);
     PushW ((unsigned short)OpAddress);
 }
 
+
+//PHA
+static void Op48E1 (void)
+{
+    PushBE (Registers.AL);
+#ifndef SA1_OPCODES
+    CPU.Cycles += ONE_CYCLE;
+#endif
+}
+
 static void Op48M1 (void)
 {
     PushB (Registers.AL);
@@ -2929,6 +2992,14 @@ static void Op48M0 (void)
 #endif
 }
 
+//PHB
+static void Op8BE1 (void)
+{
+    PushBE (Registers.DB);
+#ifndef SA1_OPCODES
+    CPU.Cycles += ONE_CYCLE;
+#endif
+}
 static void Op8B (void)
 {
     PushB (Registers.DB);
@@ -2937,6 +3008,15 @@ static void Op8B (void)
 #endif
 }
 
+//PHD NL
+static void Op0BE1 (void)
+{
+    PushWENew (Registers.D.W);
+#ifndef SA1_OPCODES
+    CPU.Cycles += ONE_CYCLE;
+#endif
+}
+
 static void Op0B (void)
 {
     PushW (Registers.D.W);
@@ -2945,6 +3025,15 @@ static void Op0B (void)
 #endif
 }
 
+//PHK
+static void Op4BE1 (void)
+{
+    PushBE (Registers.PB);
+#ifndef SA1_OPCODES
+    CPU.Cycles += ONE_CYCLE;
+#endif
+}
+
 static void Op4B (void)
 {
     PushB (Registers.PB);
@@ -2953,6 +3042,16 @@ static void Op4B (void)
 #endif
 }
 
+//PHP
+static void Op08E1 (void)
+{
+    S9xPackStatus ();
+    PushBE (Registers.PL);
+#ifndef SA1_OPCODES
+    CPU.Cycles += ONE_CYCLE;
+#endif
+}
+
 static void Op08 (void)
 {
     S9xPackStatus ();
@@ -2962,6 +3061,15 @@ static void Op08 (void)
 #endif
 }
 
+//PHX
+static void OpDAE1 (void)
+{
+    PushBE (Registers.XL);
+#ifndef SA1_OPCODES
+    CPU.Cycles += ONE_CYCLE;
+#endif
+}
+
 static void OpDAX1 (void)
 {
     PushB (Registers.XL);
@@ -2978,6 +3086,15 @@ static void OpDAX0 (void)
 #endif
 }
 
+//PHY
+static void Op5AE1 (void)
+{
+    PushBE (Registers.YL);
+#ifndef SA1_OPCODES
+    CPU.Cycles += ONE_CYCLE;
+#endif
+}
+
 static void Op5AX1 (void)
 {
     PushB (Registers.YL);
@@ -2997,12 +3114,43 @@ static void Op5AX0 (void)
 
 /* PULL Instructions ************************************************************************* */
 #define PullW(w) \
-	w = S9xGetWord (Registers.S.W + 1); \
+	w = S9xGetByte (++Registers.S.W); \
+	w |= (S9xGetByte (++Registers.S.W)<<8);
+
+/*	w = S9xGetWord (Registers.S.W + 1); \
 	Registers.S.W += 2;
+*/
 
 #define PullB(b)\
 	b = S9xGetByte (++Registers.S.W);
 
+#define PullBE(b)\
+	Registers.S.W++;\
+	Registers.SH=0x01;\
+	b = S9xGetByte (Registers.S.W);
+
+#define PullWE(w) \
+	Registers.S.W++;\
+	Registers.SH=0x01;\
+	w = S9xGetByte (Registers.S.W); \
+	Registers.S.W++; \
+	Registers.SH=0x01;\
+	w |= (S9xGetByte (Registers.S.W)<<8);
+
+#define PullWENew(w) \
+	PullW(w);\
+	Registers.SH=0x01;	
+
+//PLA
+static void Op68E1 (void)
+{
+#ifndef SA1_OPCODES
+    CPU.Cycles += TWO_CYCLES;
+#endif
+    PullBE (Registers.AL);
+    SetZN8 (Registers.AL);
+}
+
 static void Op68M1 (void)
 {
 #ifndef SA1_OPCODES
@@ -3021,6 +3169,17 @@ static void Op68M0 (void)
     SetZN16 (Registers.A.W);
 }
 
+//PLB
+static void OpABE1 (void)
+{
+#ifndef SA1_OPCODES
+    CPU.Cycles += TWO_CYCLES;
+#endif
+    PullBE (Registers.DB);
+    SetZN8 (Registers.DB);
+    ICPU.ShiftedDB = Registers.DB << 16;
+}
+
 static void OpAB (void)
 {
 #ifndef SA1_OPCODES
@@ -3032,6 +3191,16 @@ static void OpAB (void)
 }
 
 /* PHP */
+//PLD NL
+static void Op2BE1 (void)
+{
+#ifndef SA1_OPCODES
+    CPU.Cycles += TWO_CYCLES;
+#endif
+    PullWENew (Registers.D.W);
+    SetZN16 (Registers.D.W);
+}
+
 static void Op2B (void)
 {
 #ifndef SA1_OPCODES
@@ -3042,6 +3211,23 @@ static void Op2B (void)
 }
 
 /* PLP */
+static void Op28E1 (void)
+{
+#ifndef SA1_OPCODES
+    CPU.Cycles += TWO_CYCLES;
+#endif
+    PullBE (Registers.PL);
+    S9xUnpackStatus ();
+
+    if (CheckIndex ())
+    {
+	Registers.XH = 0;
+	Registers.YH = 0;
+    }
+    S9xFixCycles();
+/*     CHECK_FOR_IRQ();*/
+}
+
 static void Op28 (void)
 {
 #ifndef SA1_OPCODES
@@ -3059,6 +3245,16 @@ static void Op28 (void)
 /*     CHECK_FOR_IRQ();*/
 }
 
+//PLX
+static void OpFAE1 (void)
+{
+#ifndef SA1_OPCODES
+    CPU.Cycles += TWO_CYCLES;
+#endif
+    PullBE (Registers.XL);
+    SetZN8 (Registers.XL);
+}
+
 static void OpFAX1 (void)
 {
 #ifndef SA1_OPCODES
@@ -3077,6 +3273,16 @@ static void OpFAX0 (void)
     SetZN16 (Registers.X.W);
 }
 
+//PLY
+static void Op7AE1 (void)
+{
+#ifndef SA1_OPCODES
+    CPU.Cycles += TWO_CYCLES;
+#endif
+    PullBE (Registers.YL);
+    SetZN8 (Registers.YL);
+}
+
 static void Op7AX1 (void)
 {
 #ifndef SA1_OPCODES
@@ -3604,6 +3810,16 @@ static void Op7C (void)
 /**********************************************************************************************/
 
 /* JSL/RTL *********************************************************************************** */
+static void Op22E1 (void)
+{
+    AbsoluteLong (JUMP);
+    PushB (Registers.PB);
+    PushWENew (CPU.PC - CPU.PCBase - 1);
+    Registers.PB = (uint8) (OpAddress >> 16);
+    ICPU.ShiftedPB = OpAddress & 0xff0000;
+    S9xSetPCBase (OpAddress);
+}
+
 static void Op22 (void)
 {
     AbsoluteLong (JUMP);
@@ -3614,6 +3830,17 @@ static void Op22 (void)
     S9xSetPCBase (OpAddress);
 }
 
+static void Op6BE1 (void)
+{
+    PullWENew (Registers.PC);
+    PullB (Registers.PB);
+    ICPU.ShiftedPB = Registers.PB << 16;
+    S9xSetPCBase (ICPU.ShiftedPB + ((Registers.PC + 1) & 0xffff));
+#ifndef SA1_OPCODES
+    CPU.Cycles += TWO_CYCLES;
+#endif
+}
+
 static void Op6B (void)
 {
     PullW (Registers.PC);
@@ -3637,6 +3864,17 @@ static void Op20 (void)
 #endif
 }
 
+//JSR a,x
+static void OpFCE1 (void)
+{
+    AbsoluteIndexedIndirect (JUMP);
+    PushWENew (CPU.PC - CPU.PCBase - 1);
+    S9xSetPCBase (ICPU.ShiftedPB + OpAddress);
+#ifndef SA1_OPCODES
+    CPU.Cycles += ONE_CYCLE;
+#endif
+}
+
 static void OpFC (void)
 {
     AbsoluteIndexedIndirect (JUMP);
@@ -3881,12 +4119,15 @@ static void OpCB (void)
     else
 #endif
     {
+		if (DSP1.version == 3) return;
+
 	CPU.WaitingForInterrupt = TRUE;
 	CPU.PC--;
 #ifdef CPU_SHUTDOWN
 	if (Settings.Shutdown)
 	{
 	    CPU.Cycles = CPU.NextEvent;
+		S9xUpdateAPUTimer();
 	    if (IAPU.APUExecuting)
 	    {
 		ICPU.CPUExecuting = FALSE;
@@ -3981,6 +4222,62 @@ struct SOpcodes S9xOpcodesM1X1[256] =
     {OpFFM1}
 };
 
+struct SOpcodes S9xOpcodesE1[256] =
+{
+    {Op00},	 {Op01M1},    {Op02},      {Op03M1},    {Op04M1},
+    {Op05M1},    {Op06M1},    {Op07M1},    {Op08E1},      {Op09M1},
+    {Op0AM1},    {Op0BE1},      {Op0CM1},    {Op0DM1},    {Op0EM1},
+    {Op0FM1},    {Op10},      {Op11M1},    {Op12M1},    {Op13M1},
+    {Op14M1},    {Op15M1},    {Op16M1},    {Op17M1},    {Op18},
+    {Op19M1},    {Op1AM1},    {Op1B},      {Op1CM1},    {Op1DM1},
+    {Op1EM1},    {Op1FM1},    {Op20},      {Op21M1},    {Op22E1},
+    {Op23M1},    {Op24M1},    {Op25M1},    {Op26M1},    {Op27M1},
+    {Op28},      {Op29M1},    {Op2AM1},    {Op2BE1},      {Op2CM1},
+    {Op2DM1},    {Op2EM1},    {Op2FM1},    {Op30},      {Op31M1},
+    {Op32M1},    {Op33M1},    {Op34M1},    {Op35M1},    {Op36M1},
+    {Op37M1},    {Op38},      {Op39M1},    {Op3AM1},    {Op3B},
+    {Op3CM1},    {Op3DM1},    {Op3EM1},    {Op3FM1},    {Op40},
+    {Op41M1},    {Op42},      {Op43M1},    {Op44X1},    {Op45M1},
+    {Op46M1},    {Op47M1},    {Op48E1},    {Op49M1},    {Op4AM1},
+    {Op4BE1},      {Op4C},      {Op4DM1},    {Op4EM1},    {Op4FM1},
+    {Op50},      {Op51M1},    {Op52M1},    {Op53M1},    {Op54X1},
+    {Op55M1},    {Op56M1},    {Op57M1},    {Op58},      {Op59M1},
+    {Op5AE1},    {Op5B},      {Op5C},      {Op5DM1},    {Op5EM1},
+    {Op5FM1},    {Op60},      {Op61M1},    {Op62E1},      {Op63M1},
+    {Op64M1},    {Op65M1},    {Op66M1},    {Op67M1},    {Op68E1},
+    {Op69M1},    {Op6AM1},    {Op6BE1},      {Op6C},      {Op6DM1},
+    {Op6EM1},    {Op6FM1},    {Op70},      {Op71M1},    {Op72M1},
+    {Op73M1},    {Op74M1},    {Op75M1},    {Op76M1},    {Op77M1},
+    {Op78},      {Op79M1},    {Op7AE1},    {Op7B},      {Op7C},
+    {Op7DM1},    {Op7EM1},    {Op7FM1},    {Op80},      {Op81M1},
+    {Op82},      {Op83M1},    {Op84X1},    {Op85M1},    {Op86X1},
+    {Op87M1},    {Op88X1},    {Op89M1},    {Op8AM1},    {Op8BE1},
+    {Op8CX1},    {Op8DM1},    {Op8EX1},    {Op8FM1},    {Op90},
+    {Op91M1},    {Op92M1},    {Op93M1},    {Op94X1},    {Op95M1},
+    {Op96X1},    {Op97M1},    {Op98M1},    {Op99M1},    {Op9A},
+    {Op9BX1},    {Op9CM1},    {Op9DM1},    {Op9EM1},    {Op9FM1},
+    {OpA0X1},    {OpA1M1},    {OpA2X1},    {OpA3M1},    {OpA4X1},
+    {OpA5M1},    {OpA6X1},    {OpA7M1},    {OpA8X1},    {OpA9M1},
+    {OpAAX1},    {OpABE1},      {OpACX1},    {OpADM1},    {OpAEX1},
+    {OpAFM1},    {OpB0},      {OpB1M1},    {OpB2M1},    {OpB3M1},
+    {OpB4X1},    {OpB5M1},    {OpB6X1},    {OpB7M1},    {OpB8},
+    {OpB9M1},    {OpBAX1},    {OpBBX1},    {OpBCX1},    {OpBDM1},
+    {OpBEX1},    {OpBFM1},    {OpC0X1},    {OpC1M1},    {OpC2},
+    {OpC3M1},    {OpC4X1},    {OpC5M1},    {OpC6M1},    {OpC7M1},
+    {OpC8X1},    {OpC9M1},    {OpCAX1},    {OpCB},      {OpCCX1},
+    {OpCDM1},    {OpCEM1},    {OpCFM1},    {OpD0},      {OpD1M1},
+    {OpD2M1},    {OpD3M1},    {OpD4E1},      {OpD5M1},    {OpD6M1},
+    {OpD7M1},    {OpD8},      {OpD9M1},    {OpDAE1},    {OpDB},
+    {OpDC},      {OpDDM1},    {OpDEM1},    {OpDFM1},    {OpE0X1},
+    {OpE1M1},    {OpE2},      {OpE3M1},    {OpE4X1},    {OpE5M1},
+    {OpE6M1},    {OpE7M1},    {OpE8X1},    {OpE9M1},    {OpEA},
+    {OpEB},      {OpECX1},    {OpEDM1},    {OpEEM1},    {OpEFM1},
+    {OpF0},      {OpF1M1},    {OpF2M1},    {OpF3M1},    {OpF4E1},
+    {OpF5M1},    {OpF6M1},    {OpF7M1},    {OpF8},      {OpF9M1},
+    {OpFAE1},    {OpFB},      {OpFCE1},      {OpFDM1},    {OpFEM1},
+    {OpFFM1}
+};
+
 struct SOpcodes S9xOpcodesM1X0[256] =
 {
     {Op00},	 {Op01M1},    {Op02},      {Op03M1},    {Op04M1},
@@ -4148,3 +4445,4 @@ struct SOpcodes S9xOpcodesM0X1[256] =
     {OpFAX1},    {OpFB},      {OpFC},      {OpFDM0},    {OpFEM0},
     {OpFFM0}
 };
+
