@@ -1,50 +1,64 @@
 /*******************************************************************************
   Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  
-  (c) Copyright 1996 - 2003 Gary Henderson (gary.henderson@ntlworld.com) and
+  (c) Copyright 1996 - 2002 Gary Henderson (gary.henderson@ntlworld.com) and
                             Jerremy Koot (jkoot@snes9x.com)
 
-  (c) Copyright 2002 - 2003 Matthew Kendora and
-                            Brad Jorsch (anomie@users.sourceforge.net)
- 
+  (c) Copyright 2001 - 2004 John Weidman (jweidman@slip.net)
+
+  (c) Copyright 2002 - 2004 Brad Jorsch (anomie@users.sourceforge.net),
+                            funkyass (funkyass@spam.shaw.ca),
+                            Joel Yliluoma (http://iki.fi/bisqwit/)
+                            Kris Bleakley (codeviolation@hotmail.com),
+                            Matthew Kendora,
+                            Nach (n-a-c-h@users.sourceforge.net),
+                            Peter Bortas (peter@bortas.org) and
+                            zones (kasumitokoduck@yahoo.com)
 
-                      
   C4 x86 assembler and some C emulation code
   (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
-                            _Demo_ (_demo_@zsnes.com), and
-                            Nach (n-a-c-h@users.sourceforge.net)
-                                          
+                            _Demo_ (_demo_@zsnes.com), and Nach
+
   C4 C++ code
   (c) Copyright 2003 Brad Jorsch
 
   DSP-1 emulator code
-  (c) Copyright 1998 - 2003 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
-                            John Weidman (jweidman@slip.net),
-                            neviksti (neviksti@hotmail.com), and
-                            Kris Bleakley (stinkfish@bigpond.com)
- 
+  (c) Copyright 1998 - 2004 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
+                            John Weidman, neviksti (neviksti@hotmail.com),
+                            Kris Bleakley, Andreas Naive
+
   DSP-2 emulator code
   (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                      Lord Nightmare (lord_nightmare@users.sourceforge.net
 
   OBC1 emulator code
-  (c) Copyright 2001 - 2003 zsKnight, pagefault (pagefault@zsnes.com)
+  (c) Copyright 2001 - 2004 zsKnight, pagefault (pagefault@zsnes.com) and
+                            Kris Bleakley
   Ported from x86 assembler to C by sanmaiwashi
 
   SPC7110 and RTC C++ emulator code
   (c) Copyright 2002 Matthew Kendora with research by
                      zsKnight, John Weidman, and Dark Force
 
+  S-DD1 C emulator code
+  (c) Copyright 2003 Brad Jorsch with research by
+                     Andreas Naive and John Weidman
+ 
   S-RTC C emulator code
   (c) Copyright 2001 John Weidman
   
+  ST010 C++ emulator code
+  (c) Copyright 2003 Feather, Kris Bleakley, John Weidman and Matthew Kendora
+
   Super FX x86 assembler emulator code 
   (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 
 
   Super FX C emulator code 
-  (c) Copyright 1997 - 1999 Ivar and Gary Henderson.
+  (c) Copyright 1997 - 1999 Ivar, Gary Henderson and John Weidman
 
 
+  SH assembler code partly based on x86 assembler code
+  (c) Copyright 2002 - 2004 Marcus Comstedt (marcus@mc.pp.se) 
 
  
   Specific ports contains the works of other authors. See headers in
@@ -113,9 +127,6 @@ INLINE uint8 S9xGetByte (uint32 Address)
     case CMemory::MAP_CPU:
 		return (S9xGetCPU (Address & 0xffff));
     case CMemory::MAP_DSP:
-#ifdef DSP_DUMMY_LOOPS
-		printf("Get DSP Byte @ %06X\n", Address);
-#endif
 		return (S9xGetDSP (Address & 0xffff));
     case CMemory::MAP_SA1RAM:
     case CMemory::MAP_LOROM_SRAM:
@@ -401,11 +412,11 @@ INLINE void S9xSetByte (uint8 Byte, uint32 Address)
 		return;
 	
 	case CMemory::MAP_SETA_DSP:
-		S9xSetSetaDSP(Address, Byte);
+		S9xSetSetaDSP(Byte,Address);
 		return;
 	
 	case CMemory::MAP_SETA_RISC:
-		S9xSetST018(Address, Byte);
+		S9xSetST018(Byte,Address);
 		return;
     default:
     case CMemory::MAP_NONE:
@@ -556,13 +567,13 @@ INLINE void S9xSetWord (uint16 Word, uint32 Address)
 		return;
 	
 	case CMemory::MAP_SETA_DSP:
-		S9xSetSetaDSP (Address, Word & 0xff);
-		S9xSetSetaDSP ((Address + 1),(uint8) (Word >> 8));
+		S9xSetSetaDSP (Word & 0xff, Address);
+		S9xSetSetaDSP ((uint8) (Word >> 8),(Address + 1));
 		return;
 	
 	case CMemory::MAP_SETA_RISC:
-		S9xSetST018 (Address, Word & 0xff);
-		S9xSetST018 ((Address + 1) ,(uint8) (Word >> 8));
+		S9xSetST018 (Word & 0xff, Address);
+		S9xSetST018 ((uint8) (Word >> 8),(Address + 1));
 		return;
 
     default:
@@ -680,7 +691,7 @@ INLINE uint8 *S9xGetMemPointer (uint32 Address)
 	case CMemory::MAP_OBC_RAM:
 		return GetMemPointerOBC1(Address);
 	case CMemory::MAP_SETA_DSP:
-		return Memory.SRAM;
+		return Memory.SRAM+ ((Address & 0xffff) & Memory.SRAMMask);
     case CMemory::MAP_DEBUG:
 #ifdef DEBUGGER
 		printf ("GMP %06x\n", Address);
@@ -768,3 +779,4 @@ INLINE void S9xSetPCBase (uint32 Address)
     }
 }
 #endif
+
