@@ -1,50 +1,64 @@
 /*******************************************************************************
   Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  
-  (c) Copyright 1996 - 2003 Gary Henderson (gary.henderson@ntlworld.com) and
+  (c) Copyright 1996 - 2002 Gary Henderson (gary.henderson@ntlworld.com) and
                             Jerremy Koot (jkoot@snes9x.com)
 
-  (c) Copyright 2002 - 2003 Matthew Kendora and
-                            Brad Jorsch (anomie@users.sourceforge.net)
- 
+  (c) Copyright 2001 - 2004 John Weidman (jweidman@slip.net)
+
+  (c) Copyright 2002 - 2004 Brad Jorsch (anomie@users.sourceforge.net),
+                            funkyass (funkyass@spam.shaw.ca),
+                            Joel Yliluoma (http://iki.fi/bisqwit/)
+                            Kris Bleakley (codeviolation@hotmail.com),
+                            Matthew Kendora,
+                            Nach (n-a-c-h@users.sourceforge.net),
+                            Peter Bortas (peter@bortas.org) and
+                            zones (kasumitokoduck@yahoo.com)
 
-                      
   C4 x86 assembler and some C emulation code
   (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
-                            _Demo_ (_demo_@zsnes.com), and
-                            Nach (n-a-c-h@users.sourceforge.net)
-                                          
+                            _Demo_ (_demo_@zsnes.com), and Nach
+
   C4 C++ code
   (c) Copyright 2003 Brad Jorsch
 
   DSP-1 emulator code
-  (c) Copyright 1998 - 2003 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
-                            John Weidman (jweidman@slip.net),
-                            neviksti (neviksti@hotmail.com), and
-                            Kris Bleakley (stinkfish@bigpond.com)
- 
+  (c) Copyright 1998 - 2004 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
+                            John Weidman, neviksti (neviksti@hotmail.com),
+                            Kris Bleakley, Andreas Naive
+
   DSP-2 emulator code
   (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                      Lord Nightmare (lord_nightmare@users.sourceforge.net
 
   OBC1 emulator code
-  (c) Copyright 2001 - 2003 zsKnight, pagefault (pagefault@zsnes.com)
+  (c) Copyright 2001 - 2004 zsKnight, pagefault (pagefault@zsnes.com) and
+                            Kris Bleakley
   Ported from x86 assembler to C by sanmaiwashi
 
   SPC7110 and RTC C++ emulator code
   (c) Copyright 2002 Matthew Kendora with research by
                      zsKnight, John Weidman, and Dark Force
 
+  S-DD1 C emulator code
+  (c) Copyright 2003 Brad Jorsch with research by
+                     Andreas Naive and John Weidman
+ 
   S-RTC C emulator code
   (c) Copyright 2001 John Weidman
   
+  ST010 C++ emulator code
+  (c) Copyright 2003 Feather, Kris Bleakley, John Weidman and Matthew Kendora
+
   Super FX x86 assembler emulator code 
   (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 
 
   Super FX C emulator code 
-  (c) Copyright 1997 - 1999 Ivar and Gary Henderson.
+  (c) Copyright 1997 - 1999 Ivar, Gary Henderson and John Weidman
 
 
+  SH assembler code partly based on x86 assembler code
+  (c) Copyright 2002 - 2004 Marcus Comstedt (marcus@mc.pp.se) 
 
  
   Specific ports contains the works of other authors. See headers in
@@ -87,6 +101,7 @@
 #include "sdd1.h"
 #include "srtc.h"
 #include "spc7110.h"
+#include "movie.h"
 
 #ifndef ZSNES_FX
 #include "fxemu.h"
@@ -1011,6 +1026,7 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
 						break;
 					  case 0x3038:
 						Memory.FillRAM [Address] = Byte;
+						fx_dirtySCBR();
 						break;
 					  case 0x3039:
 						Memory.FillRAM [Address] = Byte;
@@ -1020,6 +1036,10 @@ void S9xSetPPU (uint8 Byte, uint16 Address)
 						break;
 					  case 0x303b:
 						break;
+					  case 0x303c:
+						Memory.FillRAM [Address] = Byte;
+						fx_updateRamBank(Byte);
+						break;
 					  case 0x303f:
 						Memory.FillRAM [Address] = Byte;
 						break;
@@ -1346,14 +1366,14 @@ uint8 S9xGetPPU (uint16 Address)
 	    FLUSH_REDRAW ();
 
 	    //so far, 5c77 version is always 1.
-		return (PPU.OpenBus1 = (MAX_5C77_VERSION | PPU.RangeTimeOver));
+		return (PPU.OpenBus1 = (Model->_5C77 | PPU.RangeTimeOver));
 
 	case 0x213F:
 	    // NTSC/PAL and which field flags
 	    PPU.VBeamFlip = PPU.HBeamFlip = 0;
             //neviksti found a 2 and a 3 here. SNEeSe uses a 3.
             //XXX: field flags not emulated
-	    return ((Settings.PAL ? 0x10 : 0) | (Memory.FillRAM[0x213f] & 0xc0)| MAX_5C78_VERSION) | (~PPU.OpenBus2 & 0x20);
+	    return ((Settings.PAL ? 0x10 : 0) | (Memory.FillRAM[0x213f] & 0xc0)| Model->_5C78) | (~PPU.OpenBus2 & 0x20);
 
 	case 0x2140: case 0x2141: case 0x2142: case 0x2143:
 	case 0x2144: case 0x2145: case 0x2146: case 0x2147:
@@ -1453,13 +1473,16 @@ uint8 S9xGetPPU (uint16 Address)
 	    switch (Address)
 	    {
 	    case 0x21c2:
+			if(Model->_5C77 ==2)
+		        return (0x20);
+
 			//	fprintf(stderr, "Read from $21c2!\n");
 				return OpenBus;
-	        return (0x20);
 	    case 0x21c3:
+				if(Model->_5C77 ==2)
+			        return (0);
 			//	fprintf(stderr, "Read from $21c3!\n");
 				return OpenBus;
-	        return (0);
 	    case 0x2800:
 		// For Dai Kaijyu Monogatari II
 		if (Settings.SRTC)
@@ -1765,7 +1788,7 @@ void S9xSetCPU (uint8 byte, uint16 Address)
 			break;
 		  case 0x4210:
 			// NMI ocurred flag (reset on read or write)
-			Memory.FillRAM[0x4210] = MAX_5A22_VERSION;
+			Memory.FillRAM[0x4210] = Model->_5A22;
 			return;
 		  case 0x4211:
 			// IRQ ocurred flag (reset on read or write)
@@ -2145,7 +2168,7 @@ uint8 S9xGetCPU (uint16 Address)
 			return OpenBus;
 
 		}
-		return (Memory.FillRAM [Address]);
+//		return (Memory.FillRAM [Address]);
 	}
     else
     switch (Address)
@@ -2173,9 +2196,9 @@ uint8 S9xGetCPU (uint16 Address)
 		CPU.WaitAddress = CPU.PCAtOpcodeStart;
 #endif	
 		byte = Memory.FillRAM[0x4210];
-		Memory.FillRAM[0x4210] = MAX_5A22_VERSION;
+		Memory.FillRAM[0x4210] = Model->_5A22;
 		//SNEeSe returns 2 for 5A22 version.
-		return ((byte&0x80)|(OpenBus&0x70)|MAX_5A22_VERSION);
+		return ((byte&0x80)|(OpenBus&0x70)|Model->_5A22);
 
 	  case 0x4211:
 		byte = (CPU.IRQActive & (PPU_V_BEAM_IRQ_SOURCE | PPU_H_BEAM_IRQ_SOURCE)) ? 0x80 : 0;
@@ -2380,7 +2403,7 @@ uint8 S9xGetCPU (uint16 Address)
 
 		return OpenBus;
     }
-    return (Memory.FillRAM[Address]);
+//    return (Memory.FillRAM[Address]);
 }
 
 void S9xResetPPU ()
@@ -2556,7 +2579,19 @@ void S9xResetPPU ()
 		S9xProcessMouse (1);
 	}
 	for (c = 0; c < 0x8000; c += 0x100)
-		memset (&Memory.FillRAM [c], c >> 8, 0x100);
+	{
+		if ( !Settings.SuperFX )
+		{
+			memset (&Memory.FillRAM [c], c >> 8, 0x100);
+		}
+		else if ( (unsigned)c < 0x3000 || (unsigned)c >= 0x3300 )
+		{
+			/* Don't overwrite SFX pvRegisters at 0x3000-0x32FF,
+			 * they were set in FxReset.
+			 */
+			memset (&Memory.FillRAM [c], c >> 8, 0x100);
+		}
+	}
 
 	ZeroMemory (&Memory.FillRAM [0x2100], 0x100);
 	ZeroMemory (&Memory.FillRAM [0x4200], 0x100);
@@ -2845,7 +2880,7 @@ void ProcessSuperScope ()
 			PPU.VBeamPosLatched = (uint16) (y + 1);
 			PPU.HBeamPosLatched = (uint16) x;
 			PPU.HVBeamCounterLatched = TRUE;
-			Memory.FillRAM [0x213F] |= 0x40 | MAX_5C78_VERSION;
+			Memory.FillRAM [0x213F] |= 0x40 | Model->_5C78;
 		}
 		IPPU.Joypads [1] = scope;
 	}
@@ -2942,7 +2977,7 @@ void S9xUpdateJustifiers()
 		{
 			
 			PPU.HVBeamCounterLatched = FALSE;
-			Memory.FillRAM [0x213F] = MAX_5C78_VERSION;
+			Memory.FillRAM [0x213F] = Model->_5C78;
 			
 			//process latch as Justifier 2
 			if(Settings.SecondJustifier)
@@ -2955,7 +2990,7 @@ void S9xUpdateJustifiers()
 						PPU.VBeamPosLatched = (uint16) (y + 1);
 						PPU.HBeamPosLatched = (uint16) x;
 						PPU.HVBeamCounterLatched = TRUE;
-						Memory.FillRAM [0x213F] |= 0x40 | MAX_5C78_VERSION;
+						Memory.FillRAM [0x213F] |= 0x40 | Model->_5C78;
 					}
 				}
 			}
@@ -2964,7 +2999,7 @@ void S9xUpdateJustifiers()
 		{
 			
 			PPU.HVBeamCounterLatched = FALSE;
-			Memory.FillRAM [0x213F] = MAX_5C78_VERSION;
+			Memory.FillRAM [0x213F] = Model->_5C78;
 			
 			//emulate player 1.
 			if(IPPU.Controller==SNES_JUSTIFIER)
@@ -2974,7 +3009,7 @@ void S9xUpdateJustifiers()
 					PPU.VBeamPosLatched = (uint16) (y + 1);
 					PPU.HBeamPosLatched = (uint16) x;
 					PPU.HVBeamCounterLatched = TRUE;
-					Memory.FillRAM [0x213F] |= 0x40 | MAX_5C78_VERSION;
+					Memory.FillRAM [0x213F] |= 0x40 | Model->_5C78;
 				}
 			}
 		}
@@ -2988,19 +3023,19 @@ void S9xUpdateJustifiers()
 				PPU.VBeamPosLatched = (uint16) (y + 1);
 				PPU.HBeamPosLatched = (uint16) x;
 				PPU.HVBeamCounterLatched = TRUE;
-				Memory.FillRAM [0x213F] |= 0x40 | MAX_5C78_VERSION;
+				Memory.FillRAM [0x213F] |= 0x40 | Model->_5C78;
 			}
 			else
 			{
 				PPU.HVBeamCounterLatched = FALSE;
-				Memory.FillRAM [0x213F] = MAX_5C78_VERSION;
+				Memory.FillRAM [0x213F] = Model->_5C78;
 				
 			}
 		}
 		else
 		{
 			PPU.HVBeamCounterLatched = FALSE;
-			Memory.FillRAM [0x213F] = MAX_5C78_VERSION;
+			Memory.FillRAM [0x213F] = Model->_5C78;
 		}
 	}
 }
@@ -3009,15 +3044,20 @@ void S9xUpdateJoypads ()
 {
     int i;
 
-    for (i = 0; i < 5; i++)
+	for (i = 0; i < 5; i++)
 	{
 		IPPU.Joypads [i] = S9xReadJoypad (i);
+	}
+
+	S9xMovieUpdate();
+
+	for (i = 0; i < 5; i++)
+	{
 		if (IPPU.Joypads [i] & SNES_LEFT_MASK)
 			IPPU.Joypads [i] &= ~SNES_RIGHT_MASK;
 		if (IPPU.Joypads [i] & SNES_UP_MASK)
 			IPPU.Joypads [i] &= ~SNES_DOWN_MASK;
 	}
-
 	
     // BJ: This is correct behavior AFAICT (used to be Touhaiden hack)
     if (IPPU.Controller == SNES_JOYPAD || IPPU.Controller == SNES_MULTIPLAYER5)
@@ -3256,3 +3296,4 @@ printf ("%06x: %d\n", t, FxEmulate (2000000));
 #endif
 }
 #endif
+
