@@ -1,50 +1,64 @@
 /*******************************************************************************
   Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  
-  (c) Copyright 1996 - 2003 Gary Henderson (gary.henderson@ntlworld.com) and
+  (c) Copyright 1996 - 2002 Gary Henderson (gary.henderson@ntlworld.com) and
                             Jerremy Koot (jkoot@snes9x.com)
 
-  (c) Copyright 2002 - 2003 Matthew Kendora and
-                            Brad Jorsch (anomie@users.sourceforge.net)
- 
+  (c) Copyright 2001 - 2004 John Weidman (jweidman@slip.net)
+
+  (c) Copyright 2002 - 2004 Brad Jorsch (anomie@users.sourceforge.net),
+                            funkyass (funkyass@spam.shaw.ca),
+                            Joel Yliluoma (http://iki.fi/bisqwit/)
+                            Kris Bleakley (codeviolation@hotmail.com),
+                            Matthew Kendora,
+                            Nach (n-a-c-h@users.sourceforge.net),
+                            Peter Bortas (peter@bortas.org) and
+                            zones (kasumitokoduck@yahoo.com)
 
-                      
   C4 x86 assembler and some C emulation code
   (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
-                            _Demo_ (_demo_@zsnes.com), and
-                            Nach (n-a-c-h@users.sourceforge.net)
-                                          
+                            _Demo_ (_demo_@zsnes.com), and Nach
+
   C4 C++ code
   (c) Copyright 2003 Brad Jorsch
 
   DSP-1 emulator code
-  (c) Copyright 1998 - 2003 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
-                            John Weidman (jweidman@slip.net),
-                            neviksti (neviksti@hotmail.com), and
-                            Kris Bleakley (stinkfish@bigpond.com)
- 
+  (c) Copyright 1998 - 2004 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
+                            John Weidman, neviksti (neviksti@hotmail.com),
+                            Kris Bleakley, Andreas Naive
+
   DSP-2 emulator code
   (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                      Lord Nightmare (lord_nightmare@users.sourceforge.net
 
   OBC1 emulator code
-  (c) Copyright 2001 - 2003 zsKnight, pagefault (pagefault@zsnes.com)
+  (c) Copyright 2001 - 2004 zsKnight, pagefault (pagefault@zsnes.com) and
+                            Kris Bleakley
   Ported from x86 assembler to C by sanmaiwashi
 
   SPC7110 and RTC C++ emulator code
   (c) Copyright 2002 Matthew Kendora with research by
                      zsKnight, John Weidman, and Dark Force
 
+  S-DD1 C emulator code
+  (c) Copyright 2003 Brad Jorsch with research by
+                     Andreas Naive and John Weidman
+ 
   S-RTC C emulator code
   (c) Copyright 2001 John Weidman
   
+  ST010 C++ emulator code
+  (c) Copyright 2003 Feather, Kris Bleakley, John Weidman and Matthew Kendora
+
   Super FX x86 assembler emulator code 
   (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 
 
   Super FX C emulator code 
-  (c) Copyright 1997 - 1999 Ivar and Gary Henderson.
+  (c) Copyright 1997 - 1999 Ivar, Gary Henderson and John Weidman
 
 
+  SH assembler code partly based on x86 assembler code
+  (c) Copyright 2002 - 2004 Marcus Comstedt (marcus@mc.pp.se) 
 
  
   Specific ports contains the works of other authors. See headers in
@@ -99,6 +113,7 @@
 #include "srtc.h"
 #include "sdd1.h"
 #include "spc7110.h"
+#include "movie.h"
 
 extern uint8 *SRAM;
 
@@ -127,6 +142,18 @@ enum {
 
 #define COUNT(ARRAY) (sizeof (ARRAY) / sizeof (ARRAY[0]))
 
+struct SnapshotMovieInfo
+{
+	uint32	MovieInputDataSize;
+};
+
+#undef OFFSET
+#define OFFSET(f) Offset(f,struct SnapshotMovieInfo *)
+
+static FreezeData SnapMovie [] = {
+    {OFFSET (MovieInputDataSize), 4, INT_V},
+};
+
 #undef OFFSET
 #define OFFSET(f) Offset(f,struct SCPUState *)
 
@@ -498,8 +525,6 @@ static FreezeData SnapS7RTC [] = {
 static char ROMFilename [_MAX_PATH];
 //static char SnapshotFilename [_MAX_PATH];
 
-static void Freeze (STREAM);
-static int Unfreeze (STREAM);
 void FreezeStruct (STREAM stream, char *name, void *base, FreezeData *fields,
 				   int num_fields);
 void FreezeBlock (STREAM stream, char *name, uint8 *block, int size);
@@ -508,6 +533,12 @@ int UnfreezeStruct (STREAM stream, char *name, void *base, FreezeData *fields,
 					int num_fields);
 int UnfreezeBlock (STREAM stream, char *name, uint8 *block, int size);
 
+int UnfreezeStructCopy (STREAM stream, char *name, uint8** block, FreezeData *fields, int num_fields);
+
+void UnfreezeStructFromCopy (void *base, FreezeData *fields, int num_fields, uint8* block);
+
+int UnfreezeBlockCopy (STREAM stream, char *name, uint8** block, int size);
+
 bool8 Snapshot (const char *filename)
 {
     return (S9xFreezeGame (filename));
@@ -519,8 +550,20 @@ bool8 S9xFreezeGame (const char *filename)
 	
     if (S9xOpenSnapshotFile (filename, FALSE, &stream))
     {
-		Freeze (stream);
+		S9xFreezeToStream (stream);
 		S9xCloseSnapshotFile (stream);
+
+		if(S9xMovieActive())
+		{
+			sprintf(String, "Movie snapshot %s", S9xBasename (filename));
+			S9xMessage (S9X_INFO, S9X_FREEZE_FILE_INFO, String);
+		}
+		else
+		{
+			sprintf(String, "Saved %s", S9xBasename (filename));
+			S9xMessage (S9X_INFO, S9X_FREEZE_FILE_INFO, String);
+		}
+
 		return (TRUE);
     }
     return (FALSE);
@@ -543,7 +586,7 @@ bool8 S9xUnfreezeGame (const char *filename)
     if (S9xOpenSnapshotFile (filename, TRUE, &snapshot))
     {
 		int result;
-		if ((result = Unfreeze (snapshot)) != SUCCESS)
+		if ((result = S9xUnfreezeFromStream (snapshot)) != SUCCESS)
 		{
 			switch (result)
 			{
@@ -555,6 +598,12 @@ bool8 S9xUnfreezeGame (const char *filename)
 				S9xMessage (S9X_ERROR, S9X_WRONG_VERSION,
 					"Incompatable Snes9x freeze file format version");
 				break;
+			case WRONG_MOVIE_SNAPSHOT:
+				S9xMessage (S9X_ERROR, S9X_WRONG_MOVIE_SNAPSHOT, MOVIE_ERR_SNAPSHOT_WRONG_MOVIE);
+				break;
+			case NOT_A_MOVIE_SNAPSHOT:
+				S9xMessage (S9X_ERROR, S9X_NOT_A_MOVIE_SNAPSHOT, MOVIE_ERR_SNAPSHOT_NOT_MOVIE);
+				break;
 			default:
 			case FILE_NOT_FOUND:
 				sprintf (String, "ROM image \"%s\" for freeze file not found",
@@ -565,13 +614,20 @@ bool8 S9xUnfreezeGame (const char *filename)
 			S9xCloseSnapshotFile (snapshot);
 			return (FALSE);
 		}
+
+		if(!S9xMovieActive())
+		{
+			sprintf(String, "Loaded %s", S9xBasename (filename));
+			S9xMessage (S9X_INFO, S9X_FREEZE_FILE_INFO, String);
+		}
+
 		S9xCloseSnapshotFile (snapshot);
 		return (TRUE);
     }
     return (FALSE);
 }
 
-static void Freeze (STREAM stream)
+void S9xFreezeToStream (STREAM stream)
 {
     char buffer [1024];
     int i;
@@ -599,7 +655,7 @@ static void Freeze (STREAM stream)
     FreezeStruct (stream, "REG", &Registers, SnapRegisters, COUNT (SnapRegisters));
     FreezeStruct (stream, "PPU", &PPU, SnapPPU, COUNT (SnapPPU));
     FreezeStruct (stream, "DMA", DMA, SnapDMA, COUNT (SnapDMA));
-	
+
 	// RAM and VRAM
     FreezeBlock (stream, "VRA", Memory.VRAM, 0x10000);
     FreezeBlock (stream, "RAM", Memory.RAM, 0x20000);
@@ -633,14 +689,30 @@ static void Freeze (STREAM stream)
 		FreezeStruct (stream, "RTC", &rtc_f9, SnapS7RTC, COUNT (SnapS7RTC));
 	}
 
-    S9xSetSoundMute (FALSE);
+	if (S9xMovieActive ())
+	{
+		uint8* movie_freeze_buf;
+		uint32 movie_freeze_size;
+
+		S9xMovieFreeze(&movie_freeze_buf, &movie_freeze_size);
+		if(movie_freeze_buf)
+		{
+			struct SnapshotMovieInfo mi;
+			mi.MovieInputDataSize = movie_freeze_size;
+			FreezeStruct (stream, "MOV", &mi, SnapMovie, COUNT (SnapMovie));
+		    FreezeBlock (stream, "MID", movie_freeze_buf, movie_freeze_size);
+			delete [] movie_freeze_buf;
+		}
+	}
+
+	S9xSetSoundMute (FALSE);
 #ifdef ZSNES_FX
-    if (Settings.SuperFX)
+	if (Settings.SuperFX)
 		S9xSuperFXPostSaveState ();
 #endif
 }
 
-static int Unfreeze (STREAM stream)
+int S9xUnfreezeFromStream (STREAM stream)
 {
     char buffer [_MAX_PATH + 1];
     char rom_filename [_MAX_PATH + 1];
@@ -665,118 +737,223 @@ static int Unfreeze (STREAM stream)
 			"Current loaded ROM image doesn't match that required by freeze-game file.");
     }
 	
-    uint32 old_flags = CPU.Flags;
-    uint32 sa1_old_flags = SA1.Flags;
-    S9xReset ();
-    S9xSetSoundMute (TRUE);
-	
-    if ((result = UnfreezeStruct (stream, "CPU", &CPU, SnapCPU, 
-		COUNT (SnapCPU))) != SUCCESS)
-		return (result);
-    Memory.FixROMSpeed ();
-    CPU.Flags |= old_flags & (DEBUG_MODE_FLAG | TRACE_FLAG |
-		SINGLE_STEP_FLAG | FRAME_ADVANCE_FLAG);
-    if ((result = UnfreezeStruct (stream, "REG", &Registers, SnapRegisters, COUNT (SnapRegisters))) != SUCCESS)
-		return (result);
-    if ((result = UnfreezeStruct (stream, "PPU", &PPU, SnapPPU, COUNT (SnapPPU))) != SUCCESS)
-		return (result);
-	
-    IPPU.ColorsChanged = TRUE;
-    IPPU.OBJChanged = TRUE;
-    CPU.InDMA = FALSE;
-    S9xFixColourBrightness ();
-    IPPU.RenderThisFrame = FALSE;
-	
-    if ((result = UnfreezeStruct (stream, "DMA", DMA, SnapDMA, 
-		COUNT (SnapDMA))) != SUCCESS)
-		return (result);
-    if ((result = UnfreezeBlock (stream, "VRA", Memory.VRAM, 0x10000)) != SUCCESS)
-		return (result);
-    if ((result = UnfreezeBlock (stream, "RAM", Memory.RAM, 0x20000)) != SUCCESS)
-		return (result);
-    if ((result = UnfreezeBlock (stream, "SRA", ::SRAM, 0x20000)) != SUCCESS)
-		return (result);
-    if ((result = UnfreezeBlock (stream, "FIL", Memory.FillRAM, 0x8000)) != SUCCESS)
-		return (result);
-    if (UnfreezeStruct (stream, "APU", &APU, SnapAPU, COUNT (SnapAPU)) == SUCCESS)
-    {
-		if ((result = UnfreezeStruct (stream, "ARE", &APURegisters, SnapAPURegisters,
-			COUNT (SnapAPURegisters))) != SUCCESS)
-			return (result);
-		if ((result = UnfreezeBlock (stream, "ARA", IAPU.RAM, 0x10000)) != SUCCESS)
-			return (result);
-		if ((result = UnfreezeStruct (stream, "SOU", &SoundData, SnapSoundData,
-			COUNT (SnapSoundData))) != SUCCESS)
-			return (result);
+// ## begin load ##
+	uint8* local_cpu = NULL;
+	uint8* local_registers = NULL;
+	uint8* local_ppu = NULL;
+	uint8* local_dma = NULL;
+	uint8* local_vram = NULL;
+	uint8* local_ram = NULL;
+	uint8* local_sram = NULL;
+	uint8* local_fillram = NULL;
+	uint8* local_apu = NULL;
+	uint8* local_apu_registers = NULL;
+	uint8* local_apu_ram = NULL;
+	uint8* local_apu_sounddata = NULL;
+	uint8* local_sa1 = NULL;
+	uint8* local_sa1_registers = NULL;
+	uint8* local_spc = NULL;
+	uint8* local_spc_rtc = NULL;
+	uint8* local_movie_data = NULL;
+
+	do
+	{
+		if ((result = UnfreezeStructCopy (stream, "CPU", &local_cpu, SnapCPU, COUNT (SnapCPU))) != SUCCESS)
+			break;
+		if ((result = UnfreezeStructCopy (stream, "REG", &local_registers, SnapRegisters, COUNT (SnapRegisters))) != SUCCESS)
+			break;
+		if ((result = UnfreezeStructCopy (stream, "PPU", &local_ppu, SnapPPU, COUNT (SnapPPU))) != SUCCESS)
+			break;
+		if ((result = UnfreezeStructCopy (stream, "DMA", &local_dma, SnapDMA, COUNT (SnapDMA))) != SUCCESS)
+			break;
+		if ((result = UnfreezeBlockCopy (stream, "VRA", &local_vram, 0x10000)) != SUCCESS)
+			break;
+		if ((result = UnfreezeBlockCopy (stream, "RAM", &local_ram, 0x20000)) != SUCCESS)
+			break;
+		if ((result = UnfreezeBlockCopy (stream, "SRA", &local_sram, 0x20000)) != SUCCESS)
+			break;
+		if ((result = UnfreezeBlockCopy (stream, "FIL", &local_fillram, 0x8000)) != SUCCESS)
+			break;
+		if (UnfreezeStructCopy (stream, "APU", &local_apu, SnapAPU, COUNT (SnapAPU)) == SUCCESS)
+		{
+			if ((result = UnfreezeStructCopy (stream, "ARE", &local_apu_registers, SnapAPURegisters, COUNT (SnapAPURegisters))) != SUCCESS)
+				break;
+			if ((result = UnfreezeBlockCopy (stream, "ARA", &local_apu_ram, 0x10000)) != SUCCESS)
+				break;
+			if ((result = UnfreezeStructCopy (stream, "SOU", &local_apu_sounddata, SnapSoundData, COUNT (SnapSoundData))) != SUCCESS)
+				break;
+		}
+		if ((result = UnfreezeStructCopy (stream, "SA1", &local_sa1, SnapSA1, COUNT(SnapSA1))) == SUCCESS)
+		{
+			if ((result = UnfreezeStructCopy (stream, "SAR", &local_sa1_registers, SnapSA1Registers, COUNT (SnapSA1Registers))) != SUCCESS)
+				break;
+		}
 		
-		S9xSetSoundMute (FALSE);
-		IAPU.PC = IAPU.RAM + APURegisters.PC;
-		S9xAPUUnpackStatus ();
-		if (APUCheckDirectPage ())
-			IAPU.DirectPage = IAPU.RAM + 0x100;
-		else
-			IAPU.DirectPage = IAPU.RAM;
-		Settings.APUEnabled = TRUE;
-		IAPU.APUExecuting = TRUE;
-    }
-    else
-    {
-		Settings.APUEnabled = FALSE;
-		IAPU.APUExecuting = FALSE;
+		if ((result = UnfreezeStructCopy (stream, "SP7", &local_spc, SnapSPC7110, COUNT(SnapSPC7110))) != SUCCESS)
+		{
+			if(Settings.SPC7110)
+				break;
+		}
+		if ((result = UnfreezeStructCopy (stream, "RTC", &local_spc_rtc, SnapS7RTC, COUNT (SnapS7RTC))) != SUCCESS)
+		{
+			if(Settings.SPC7110RTC)
+				break;
+		}
+
+		if (S9xMovieActive ())
+		{
+			SnapshotMovieInfo mi;
+			if ((result = UnfreezeStruct (stream, "MOV", &mi, SnapMovie, COUNT(SnapMovie))) != SUCCESS)
+			{
+				result = NOT_A_MOVIE_SNAPSHOT;
+				break;
+			}
+
+			if ((result = UnfreezeBlockCopy (stream, "MID", &local_movie_data, mi.MovieInputDataSize)) != SUCCESS)
+			{
+				result = NOT_A_MOVIE_SNAPSHOT;
+				break;
+			}
+
+			if (!S9xMovieUnfreeze(local_movie_data, mi.MovieInputDataSize))
+			{
+				result = WRONG_MOVIE_SNAPSHOT;
+				break;
+			}
+		}
+
+		result=SUCCESS;
+
+	} while(false);
+// ## end load ##
+
+	if (result == SUCCESS)
+	{
+		uint32 old_flags = CPU.Flags;
+		uint32 sa1_old_flags = SA1.Flags;
+		S9xReset ();
 		S9xSetSoundMute (TRUE);
-    }
-    if ((result = UnfreezeStruct (stream, "SA1", &SA1, SnapSA1, 
-		COUNT(SnapSA1))) == SUCCESS)
-    {
-		if ((result = UnfreezeStruct (stream, "SAR", &SA1Registers, 
-			SnapSA1Registers, COUNT (SnapSA1Registers))) != SUCCESS)
-			return (result);
-		
-		S9xFixSA1AfterSnapshotLoad ();
-		SA1.Flags |= sa1_old_flags & (TRACE_FLAG);
-    }
-	
-	if ((result = UnfreezeStruct (stream, "SP7", &s7r, SnapSPC7110, 
-		COUNT(SnapSPC7110))) != SUCCESS)
-    {
-		if(Settings.SPC7110)
-			return result;
-	}
-	if ((result = UnfreezeStruct (stream, "RTC", &rtc_f9, 
-			SnapS7RTC, COUNT (SnapS7RTC))) == SUCCESS)
-	{
-		S9xUpdateRTC();
-	}
-	else
-	{
-		if(Settings.SPC7110RTC)
-			return result;
-	}
-    S9xFixSoundAfterSnapshotLoad ();
 
-	if(!Memory.FillRAM[0x4213]){
-		// most likely an old savestate
-		Memory.FillRAM[0x4213]=Memory.FillRAM[0x4201];
-		if(!Memory.FillRAM[0x4213])
-			Memory.FillRAM[0x4213]=Memory.FillRAM[0x4201]=0xFF;
-	}
+		UnfreezeStructFromCopy (&CPU, SnapCPU, COUNT (SnapCPU), local_cpu);
+		UnfreezeStructFromCopy (&Registers, SnapRegisters, COUNT (SnapRegisters), local_registers);
+		UnfreezeStructFromCopy (&PPU, SnapPPU, COUNT (SnapPPU), local_ppu);
+		UnfreezeStructFromCopy (DMA, SnapDMA, COUNT (SnapDMA), local_dma);
+		memcpy (Memory.VRAM, local_vram, 0x10000);
+		memcpy (Memory.RAM, local_ram, 0x20000);
+		memcpy (::SRAM, local_sram, 0x20000);
+		memcpy (Memory.FillRAM, local_fillram, 0x8000);
+		if(local_apu)
+		{
+			UnfreezeStructFromCopy (&APU, SnapAPU, COUNT (SnapAPU), local_apu);
+			UnfreezeStructFromCopy (&APURegisters, SnapAPURegisters, COUNT (SnapAPURegisters), local_apu_registers);
+			memcpy (IAPU.RAM, local_apu_ram, 0x10000);
+			UnfreezeStructFromCopy (&SoundData, SnapSoundData, COUNT (SnapSoundData), local_apu_sounddata);
+		}
+		if(local_sa1)
+		{
+			UnfreezeStructFromCopy (&SA1, SnapSA1, COUNT (SnapSA1), local_sa1);
+			UnfreezeStructFromCopy (&SA1Registers, SnapSA1Registers, COUNT (SnapSA1Registers), local_sa1_registers);
+		}
+		if(local_spc)
+		{
+			UnfreezeStructFromCopy (&s7r, SnapSPC7110, COUNT (SnapSPC7110), local_spc);
+		}
+		if(local_spc_rtc)
+		{
+			UnfreezeStructFromCopy (&rtc_f9, SnapS7RTC, COUNT (SnapS7RTC), local_spc_rtc);
+		}
+
+		Memory.FixROMSpeed ();
+		CPU.Flags |= old_flags & (DEBUG_MODE_FLAG | TRACE_FLAG |
+			SINGLE_STEP_FLAG | FRAME_ADVANCE_FLAG);
+
+	    IPPU.ColorsChanged = TRUE;
+		IPPU.OBJChanged = TRUE;
+		CPU.InDMA = FALSE;
+		S9xFixColourBrightness ();
+		IPPU.RenderThisFrame = FALSE;
+
+		if (local_apu)
+		{
+			S9xSetSoundMute (FALSE);
+			IAPU.PC = IAPU.RAM + APURegisters.PC;
+			S9xAPUUnpackStatus ();
+			if (APUCheckDirectPage ())
+				IAPU.DirectPage = IAPU.RAM + 0x100;
+			else
+				IAPU.DirectPage = IAPU.RAM;
+			Settings.APUEnabled = TRUE;
+			IAPU.APUExecuting = TRUE;
+		}
+		else
+		{
+			Settings.APUEnabled = FALSE;
+			IAPU.APUExecuting = FALSE;
+			S9xSetSoundMute (TRUE);
+		}
+
+		if (local_sa1)
+		{
+			S9xFixSA1AfterSnapshotLoad ();
+			SA1.Flags |= sa1_old_flags & (TRACE_FLAG);
+		}
+
+		if (local_spc_rtc)
+		{
+			S9xUpdateRTC();
+		}
+
+		S9xFixSoundAfterSnapshotLoad ();
+
+		uint8 hdma_byte = Memory.FillRAM[0x420c];
+		S9xSetCPU(hdma_byte, 0x420c);
+
+		if(!Memory.FillRAM[0x4213]){
+			// most likely an old savestate
+			Memory.FillRAM[0x4213]=Memory.FillRAM[0x4201];
+			if(!Memory.FillRAM[0x4213])
+				Memory.FillRAM[0x4213]=Memory.FillRAM[0x4201]=0xFF;
+		}
+
+		ICPU.ShiftedPB = Registers.PB << 16;
+		ICPU.ShiftedDB = Registers.DB << 16;
+		S9xSetPCBase (ICPU.ShiftedPB + Registers.PC);
+		S9xUnpackStatus ();
+		S9xFixCycles ();
+//		S9xReschedule ();				// <-- this causes desync when recording or playing movies
 
-    ICPU.ShiftedPB = Registers.PB << 16;
-    ICPU.ShiftedDB = Registers.DB << 16;
-    S9xSetPCBase (ICPU.ShiftedPB + Registers.PC);
-    S9xUnpackStatus ();
-    S9xFixCycles ();
-    S9xReschedule ();
 #ifdef ZSNES_FX
-    if (Settings.SuperFX)
-		S9xSuperFXPostLoadState ();
+		if (Settings.SuperFX)
+			S9xSuperFXPostLoadState ();
 #endif
-	
-    S9xSRTCPostLoadState ();
-    if (Settings.SDD1)
-		S9xSDD1PostLoadState ();
-	
-    return (SUCCESS);
+		
+		S9xSRTCPostLoadState ();
+		if (Settings.SDD1)
+			S9xSDD1PostLoadState ();
+			
+		IAPU.NextAPUTimerPos = CPU.Cycles * 10000L;
+		IAPU.APUTimerCounter = 0; 
+	}
+
+	if (local_cpu)           delete [] local_cpu;
+	if (local_registers)     delete [] local_registers;
+	if (local_ppu)           delete [] local_ppu;
+	if (local_dma)           delete [] local_dma;
+	if (local_vram)          delete [] local_vram;
+	if (local_ram)           delete [] local_ram;
+	if (local_sram)          delete [] local_sram;
+	if (local_fillram)       delete [] local_fillram;
+	if (local_apu)           delete [] local_apu;
+	if (local_apu_registers) delete [] local_apu_registers;
+	if (local_apu_ram)       delete [] local_apu_ram;
+	if (local_apu_sounddata) delete [] local_apu_sounddata;
+	if (local_sa1)           delete [] local_sa1;
+	if (local_sa1_registers) delete [] local_sa1_registers;
+	if (local_spc)           delete [] local_spc;
+	if (local_spc_rtc)       delete [] local_spc_rtc;
+	if (local_movie_data)    delete [] local_movie_data;
+
+	return (result);
 }
 
 int FreezeSize (int size, int type)
@@ -995,7 +1172,7 @@ int UnfreezeBlock (STREAM stream, char *name, uint8 *block, int size)
 		REVERT_STREAM(stream, FIND_STREAM(stream)-11, 0);
 		return (WRONG_FORMAT);
     }
-	
+
     if (len > size)
     {
 		rem = len - size;
@@ -1016,6 +1193,109 @@ int UnfreezeBlock (STREAM stream, char *name, uint8 *block, int size)
     return (SUCCESS);
 }
 
+int UnfreezeStructCopy (STREAM stream, char *name, uint8** block, FreezeData *fields, int num_fields)
+{
+    // Work out the size of the required block
+    int len = 0;
+    int i;
+	
+    for (i = 0; i < num_fields; i++)
+    {
+		if (fields [i].offset + FreezeSize (fields [i].size, 
+			fields [i].type) > len)
+			len = fields [i].offset + FreezeSize (fields [i].size, 
+			fields [i].type);
+    }
+	
+    return (UnfreezeBlockCopy (stream, name, block, len));
+}
+
+void UnfreezeStructFromCopy (void *base, FreezeData *fields, int num_fields, uint8* block)
+{
+	int i;
+	int j;
+    uint8 *ptr = block;
+    uint16 word;
+    uint32 dword;
+    int64  qword;
+	
+    // Unpack the block of data into a C structure
+    for (i = 0; i < num_fields; i++)
+    {
+		switch (fields [i].type)
+		{
+		case INT_V:
+			switch (fields [i].size)
+			{
+			case 1:
+				*((uint8 *) base + fields [i].offset) = *ptr++;
+				break;
+			case 2:
+				word  = *ptr++ << 8;
+				word |= *ptr++;
+				*((uint16 *) ((uint8 *) base + fields [i].offset)) = word;
+				break;
+			case 4:
+				dword  = *ptr++ << 24;
+				dword |= *ptr++ << 16;
+				dword |= *ptr++ << 8;
+				dword |= *ptr++;
+				*((uint32 *) ((uint8 *) base + fields [i].offset)) = dword;
+				break;
+			case 8:
+				qword  = (int64) *ptr++ << 56;
+				qword |= (int64) *ptr++ << 48;
+				qword |= (int64) *ptr++ << 40;
+				qword |= (int64) *ptr++ << 32;
+				qword |= (int64) *ptr++ << 24;
+				qword |= (int64) *ptr++ << 16;
+				qword |= (int64) *ptr++ << 8;
+				qword |= (int64) *ptr++;
+				*((int64 *) ((uint8 *) base + fields [i].offset)) = qword;
+				break;
+			}
+			break;
+			case uint8_ARRAY_V:
+				memmove ((uint8 *) base + fields [i].offset, ptr, fields [i].size);
+				ptr += fields [i].size;
+				break;
+			case uint16_ARRAY_V:
+				for (j = 0; j < fields [i].size; j++)
+				{
+					word  = *ptr++ << 8;
+					word |= *ptr++;
+					*((uint16 *) ((uint8 *) base + fields [i].offset + j * 2)) = word;
+				}
+				break;
+			case uint32_ARRAY_V:
+				for (j = 0; j < fields [i].size; j++)
+				{
+					dword  = *ptr++ << 24;
+					dword |= *ptr++ << 16;
+					dword |= *ptr++ << 8;
+					dword |= *ptr++;
+					*((uint32 *) ((uint8 *) base + fields [i].offset + j * 4)) = dword;
+				}
+				break;
+		}
+    }
+}
+
+int UnfreezeBlockCopy (STREAM stream, char *name, uint8** block, int size)
+{
+    *block = new uint8 [size];
+    int result;
+	
+    if ((result = UnfreezeBlock (stream, name, *block, size)) != SUCCESS)
+    {
+		delete [] (*block);
+		*block = NULL;
+		return (result);
+    }
+	
+    return (result);
+}
+
 extern uint8 spc_dump_dsp[0x100];
 
 bool8 S9xSPCDump (const char *filename)
@@ -1535,3 +1815,4 @@ fread(&temp, 1, 4, fs);
     fclose (fs);
     return (FALSE);
 }
+
