@@ -1,50 +1,64 @@
 /*******************************************************************************
   Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  
-  (c) Copyright 1996 - 2003 Gary Henderson (gary.henderson@ntlworld.com) and
+  (c) Copyright 1996 - 2002 Gary Henderson (gary.henderson@ntlworld.com) and
                             Jerremy Koot (jkoot@snes9x.com)
 
-  (c) Copyright 2002 - 2003 Matthew Kendora and
-                            Brad Jorsch (anomie@users.sourceforge.net)
- 
+  (c) Copyright 2001 - 2004 John Weidman (jweidman@slip.net)
+
+  (c) Copyright 2002 - 2004 Brad Jorsch (anomie@users.sourceforge.net),
+                            funkyass (funkyass@spam.shaw.ca),
+                            Joel Yliluoma (http://iki.fi/bisqwit/)
+                            Kris Bleakley (codeviolation@hotmail.com),
+                            Matthew Kendora,
+                            Nach (n-a-c-h@users.sourceforge.net),
+                            Peter Bortas (peter@bortas.org) and
+                            zones (kasumitokoduck@yahoo.com)
 
-                      
   C4 x86 assembler and some C emulation code
   (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
-                            _Demo_ (_demo_@zsnes.com), and
-                            Nach (n-a-c-h@users.sourceforge.net)
-                                          
+                            _Demo_ (_demo_@zsnes.com), and Nach
+
   C4 C++ code
   (c) Copyright 2003 Brad Jorsch
 
   DSP-1 emulator code
-  (c) Copyright 1998 - 2003 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
-                            John Weidman (jweidman@slip.net),
-                            neviksti (neviksti@hotmail.com), and
-                            Kris Bleakley (stinkfish@bigpond.com)
- 
+  (c) Copyright 1998 - 2004 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
+                            John Weidman, neviksti (neviksti@hotmail.com),
+                            Kris Bleakley, Andreas Naive
+
   DSP-2 emulator code
   (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                      Lord Nightmare (lord_nightmare@users.sourceforge.net
 
   OBC1 emulator code
-  (c) Copyright 2001 - 2003 zsKnight, pagefault (pagefault@zsnes.com)
+  (c) Copyright 2001 - 2004 zsKnight, pagefault (pagefault@zsnes.com) and
+                            Kris Bleakley
   Ported from x86 assembler to C by sanmaiwashi
 
   SPC7110 and RTC C++ emulator code
   (c) Copyright 2002 Matthew Kendora with research by
                      zsKnight, John Weidman, and Dark Force
 
+  S-DD1 C emulator code
+  (c) Copyright 2003 Brad Jorsch with research by
+                     Andreas Naive and John Weidman
+ 
   S-RTC C emulator code
   (c) Copyright 2001 John Weidman
   
+  ST010 C++ emulator code
+  (c) Copyright 2003 Feather, Kris Bleakley, John Weidman and Matthew Kendora
+
   Super FX x86 assembler emulator code 
   (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 
 
   Super FX C emulator code 
-  (c) Copyright 1997 - 1999 Ivar and Gary Henderson.
+  (c) Copyright 1997 - 1999 Ivar, Gary Henderson and John Weidman
 
 
+  SH assembler code partly based on x86 assembler code
+  (c) Copyright 2002 - 2004 Marcus Comstedt (marcus@mc.pp.se) 
 
  
   Specific ports contains the works of other authors. See headers in
@@ -233,7 +247,7 @@ void S9xSetSoundVolume (int channel, short volume_left, short volume_right)
 
 void S9xSetMasterVolume (short volume_left, short volume_right)
 {
-    if (Settings.DisableMasterVolume)
+    if (Settings.DisableMasterVolume || SNESGameFixes.EchoOnlyOutput)
     {
 		SoundData.master_volume_left = 127;
 		SoundData.master_volume_right = 127;
@@ -786,20 +800,13 @@ void DecodeBlock (Channel *ch)
 	// Seperate out the header parts used for decoding
 
 	shift = filter >> 4;
+	
+	// Header validity check: if range(shift) is over 12, ignore
+	// all bits of the data for that block except for the sign bit of each
+	invalid_header = !(shift < 0xD);
+
 	filter = filter&0x0c;
 
-    // Header validity check: if range(shift) is over 12, ignore
-	// all bits of the data for that block except for the sign bit of each one
-	if ((filter&0xF0) < 0xD0)
-	{
-		invalid_header = false;
-	}
-	else
-	{
-		// fprintf(stderr,"invalid header! header: %02x\n", filter);
-		invalid_header = true;
-	}
-	
 	int32 prev0 = ch->previous [0];
 	int32 prev1 = ch->previous [1];
 	
@@ -839,8 +846,7 @@ void DecodeBlock (Channel *ch)
 					break;
 				
 			}
-			if (out > 32767) out = 32767;
-			else if (out < -32768) out = -32768;
+			CLIP16(out);
 				*raw++ = (signed short)(out<<1);
 			prev1=(signed short)prev0;
 			prev0=(signed short)(out<<1);
@@ -855,11 +861,8 @@ void DecodeBlock (Channel *ch)
 
 void MixStereo (int sample_count)
 {
-#if defined(TARGET_OS_MAC) && TARGET_OS_MAC
     static int wave[SOUND_BUFFER_SIZE];
-#else
-    int wave[SOUND_BUFFER_SIZE];
-#endif
+
     int pitch_mod = SoundData.pitch_mod & ~APU.DSP[APU_NON];
 	
     for (uint32 J = 0; J < NUM_CHANNELS; J++) 
@@ -889,7 +892,7 @@ void MixStereo (int sample_count)
 			ch->next_sample=ch->block[ch->sample_pointer];
 			ch->interpolate = 0;
 			  
-			   if (Settings.InterpolatedSound && freq0 < FIXED_POINT && !mod)
+			if (Settings.InterpolatedSound && freq0 < FIXED_POINT && !mod)
 			   ch->interpolate = ((ch->next_sample - ch->sample) * 
 			   (long) freq0) / (long) FIXED_POINT;
 		}
@@ -1153,11 +1156,8 @@ END_OF_FUNCTION(MixStereo);
 
 void MixMono (int sample_count)
 {
-#if defined(TARGET_OS_MAC) && TARGET_OS_MAC
     static int wave[SOUND_BUFFER_SIZE];
-#else
-    int wave[SOUND_BUFFER_SIZE];
-#endif
+
     int pitch_mod = SoundData.pitch_mod & (~APU.DSP[APU_NON]);
 	
     for (uint32 J = 0; J < NUM_CHANNELS; J++) 
@@ -1443,15 +1443,15 @@ extern uint8 int2ulaw (int);
 #endif
 
 // For backwards compatibility with older port specific code
-void S9xMixSamples (uint8 *buffer, int sample_count)
+void S9xMixSamplesO (uint8 *buffer, int sample_count, int byte_offset)
 {
-    S9xMixSamplesO (buffer, sample_count, 0);
+    S9xMixSamples (buffer+byte_offset, sample_count);
 }
 #ifdef __DJGPP
-END_OF_FUNCTION(S9xMixSamples);
+END_OF_FUNCTION(S9xMixSamplesO);
 #endif
 
-void S9xMixSamplesO (uint8 *buffer, int sample_count, int byte_offset)
+void S9xMixSamples (uint8 *buffer, int sample_count)
 {
     int J;
     int I;
@@ -1476,11 +1476,10 @@ void S9xMixSamplesO (uint8 *buffer, int sample_count, int byte_offset)
 		// 16-bit sound
 		if (so.mute_sound)
 		{
-            memset (buffer + byte_offset, 0, byte_count);
+            memset (buffer, 0, byte_count);
 		}
 		else
 		{
-			int O = byte_offset >> 1;
 			if (SoundData.echo_enable && SoundData.echo_buffer_size)
 			{
 				if (so.stereo)
@@ -1504,7 +1503,7 @@ void S9xMixSamplesO (uint8 *buffer, int sample_count, int byte_offset)
 								E * SoundData.echo_volume [J & 1]) / VOL_DIV16;
 							
 							CLIP16(I);
-							((signed short *) buffer)[J + O] = I;
+							((signed short *) buffer)[J] = I;
 						}
 					}
 					else
@@ -1537,7 +1536,7 @@ void S9xMixSamplesO (uint8 *buffer, int sample_count, int byte_offset)
 								E * SoundData.echo_volume [J & 1]) / VOL_DIV16;
 							
 							CLIP16(I);
-							((signed short *) buffer)[J + O] = I;
+							((signed short *) buffer)[J] = I;
 						}
 					}
 				}
@@ -1561,7 +1560,7 @@ void S9xMixSamplesO (uint8 *buffer, int sample_count, int byte_offset)
 								SoundData.master_volume [0] +
 								E * SoundData.echo_volume [0]) / VOL_DIV16;
 							CLIP16(I);
-							((signed short *) buffer)[J + O] = I;
+							((signed short *) buffer)[J] = I;
 						}
 					}
 					else
@@ -1592,7 +1591,7 @@ void S9xMixSamplesO (uint8 *buffer, int sample_count, int byte_offset)
 							I = (MixBuffer [J] * SoundData.master_volume [0] +
 								E * SoundData.echo_volume [0]) / VOL_DIV16;
 							CLIP16(I);
-							((signed short *) buffer)[J + O] = I;
+							((signed short *) buffer)[J] = I;
 						}
 					}
 				}
@@ -1606,19 +1605,17 @@ void S9xMixSamplesO (uint8 *buffer, int sample_count, int byte_offset)
 					SoundData.master_volume [J & 1]) / VOL_DIV16;
 				
 				CLIP16(I);
-				((signed short *) buffer)[J + O] = I;
+				((signed short *) buffer)[J] = I;
 			}
 		}
 	}
     }
     else
     {
-		int O = byte_offset;
-		
 		// 8-bit sound
 		if (so.mute_sound)
 		{
-            memset (buffer + O, 128, sample_count);
+            memset (buffer, 128, sample_count);
 		}
 		else
 #ifdef __sun
@@ -1628,7 +1625,7 @@ void S9xMixSamplesO (uint8 *buffer, int sample_count, int byte_offset)
 				{
 					I = (MixBuffer [J] * SoundData.master_volume_left) / VOL_DIV16;
 					CLIP16(I);
-					buffer[J + O] = int2ulaw (I);
+					buffer[J] = int2ulaw (I);
 				}
 			}
 			else
@@ -1656,7 +1653,7 @@ void S9xMixSamplesO (uint8 *buffer, int sample_count, int byte_offset)
 									SoundData.master_volume [J & 1] +
 									E * SoundData.echo_volume [J & 1]) / VOL_DIV8;
 								CLIP8(I);
-								buffer [J + O] = I + 128;
+								buffer [J] = I + 128;
 							}
 						}
 						else
@@ -1688,7 +1685,7 @@ void S9xMixSamplesO (uint8 *buffer, int sample_count, int byte_offset)
 									SoundData.master_volume [J & 1] +
 									E * SoundData.echo_volume [J & 1]) / VOL_DIV8;
 								CLIP8(I);
-								buffer [J + O] = I + 128;
+								buffer [J] = I + 128;
 							}
 						}
 					}
@@ -1711,7 +1708,7 @@ void S9xMixSamplesO (uint8 *buffer, int sample_count, int byte_offset)
 								I = (MixBuffer [J] * SoundData.master_volume [0] +
 									E * SoundData.echo_volume [0]) / VOL_DIV8;
 								CLIP8(I);
-								buffer [J + O] = I + 128;
+								buffer [J] = I + 128;
 							}
 						}
 						else
@@ -1742,7 +1739,7 @@ void S9xMixSamplesO (uint8 *buffer, int sample_count, int byte_offset)
 								I = (MixBuffer [J] * SoundData.master_volume [0] +
 									E * SoundData.echo_volume [0]) / VOL_DIV8;
 								CLIP8(I);
-								buffer [J + O] = I + 128;
+								buffer [J] = I + 128;
 							}
 						}
 					}
@@ -1755,7 +1752,7 @@ void S9xMixSamplesO (uint8 *buffer, int sample_count, int byte_offset)
 				I = (MixBuffer [J] * 
 					SoundData.master_volume [J & 1]) / VOL_DIV8;
 				CLIP8(I);
-				buffer [J + O] = I + 128;
+				buffer [J] = I + 128;
 			}
 		}
 	}
@@ -1763,7 +1760,7 @@ void S9xMixSamplesO (uint8 *buffer, int sample_count, int byte_offset)
 }
 
 #ifdef __DJGPP
-END_OF_FUNCTION(S9xMixSamplesO);
+END_OF_FUNCTION(S9xMixSamples);
 #endif
 
 void S9xResetSound (bool8 full)
@@ -1868,8 +1865,13 @@ bool8 S9xInitSound (int mode, bool8 stereo, int buffer_size)
     S9xSetSoundMute (TRUE);
     if (!S9xOpenSoundDevice (mode, stereo, buffer_size))
     {
+#ifdef NOSOUND
+                S9xMessage (S9X_WARNING, S9X_SOUND_NOT_BUILT,
+                            "No sound support compiled in");
+#else
 		S9xMessage (S9X_ERROR, S9X_SOUND_DEVICE_OPEN_FAILED,
-			"Sound device open failed");
+                            "Sound device open failed");
+#endif
 		return (0);
     }
 	
@@ -2024,3 +2026,4 @@ void S9xPlaySample (int channel)
 		APU.DSP [APU_ADSR2 + (channel << 4)]);
 }
 
+
