@@ -1,50 +1,64 @@
 /*******************************************************************************
   Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
  
-  (c) Copyright 1996 - 2003 Gary Henderson (gary.henderson@ntlworld.com) and
+  (c) Copyright 1996 - 2002 Gary Henderson (gary.henderson@ntlworld.com) and
                             Jerremy Koot (jkoot@snes9x.com)
 
-  (c) Copyright 2002 - 2003 Matthew Kendora and
-                            Brad Jorsch (anomie@users.sourceforge.net)
- 
+  (c) Copyright 2001 - 2004 John Weidman (jweidman@slip.net)
+
+  (c) Copyright 2002 - 2004 Brad Jorsch (anomie@users.sourceforge.net),
+                            funkyass (funkyass@spam.shaw.ca),
+                            Joel Yliluoma (http://iki.fi/bisqwit/)
+                            Kris Bleakley (codeviolation@hotmail.com),
+                            Matthew Kendora,
+                            Nach (n-a-c-h@users.sourceforge.net),
+                            Peter Bortas (peter@bortas.org) and
+                            zones (kasumitokoduck@yahoo.com)
 
-                      
   C4 x86 assembler and some C emulation code
   (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
-                            _Demo_ (_demo_@zsnes.com), and
-                            Nach (n-a-c-h@users.sourceforge.net)
-                                          
+                            _Demo_ (_demo_@zsnes.com), and Nach
+
   C4 C++ code
   (c) Copyright 2003 Brad Jorsch
 
   DSP-1 emulator code
-  (c) Copyright 1998 - 2003 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
-                            John Weidman (jweidman@slip.net),
-                            neviksti (neviksti@hotmail.com), and
-                            Kris Bleakley (stinkfish@bigpond.com)
- 
+  (c) Copyright 1998 - 2004 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
+                            John Weidman, neviksti (neviksti@hotmail.com),
+                            Kris Bleakley, Andreas Naive
+
   DSP-2 emulator code
   (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                      Lord Nightmare (lord_nightmare@users.sourceforge.net
 
   OBC1 emulator code
-  (c) Copyright 2001 - 2003 zsKnight, pagefault (pagefault@zsnes.com)
+  (c) Copyright 2001 - 2004 zsKnight, pagefault (pagefault@zsnes.com) and
+                            Kris Bleakley
   Ported from x86 assembler to C by sanmaiwashi
 
   SPC7110 and RTC C++ emulator code
   (c) Copyright 2002 Matthew Kendora with research by
                      zsKnight, John Weidman, and Dark Force
 
+  S-DD1 C emulator code
+  (c) Copyright 2003 Brad Jorsch with research by
+                     Andreas Naive and John Weidman
+ 
   S-RTC C emulator code
   (c) Copyright 2001 John Weidman
   
+  ST010 C++ emulator code
+  (c) Copyright 2003 Feather, Kris Bleakley, John Weidman and Matthew Kendora
+
   Super FX x86 assembler emulator code 
   (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 
 
   Super FX C emulator code 
-  (c) Copyright 1997 - 1999 Ivar and Gary Henderson.
+  (c) Copyright 1997 - 1999 Ivar, Gary Henderson and John Weidman
 
 
+  SH assembler code partly based on x86 assembler code
+  (c) Copyright 2002 - 2004 Marcus Comstedt (marcus@mc.pp.se) 
 
  
   Specific ports contains the works of other authors. See headers in
@@ -85,6 +99,9 @@
 #include "gfx.h"
 #include "soundux.h"
 #include "netplay.h"
+#include "movie.h"
+#include "..\AVIOutput.h"
+#include "..\language.h"
 
 #include <direct.h>
 
@@ -105,6 +122,21 @@ FILE *offsets_h = NULL;
 
 struct SJoyState Joystick [16];
 uint32 joypads [5];
+bool8 do_frame_adjust=false;
+
+// avi variables
+static uint8* avi_buffer = NULL;
+static uint8* avi_sound_buffer = NULL;
+static int avi_sound_bytes_per_sample = 0;
+static int avi_sound_samples_per_update = 0;
+static int avi_width = 0;
+static int avi_height = 0;
+static uint32 avi_skip_frames = 0;
+void Convert8To24Packed (SSurface *src, SSurface *dst, RECT *srect);
+void Convert16To24Packed (SSurface *src, SSurface *dst, RECT *srect);
+void DoAVIOpen(const char* filename);
+void DoAVIClose(int reason);
+static void DoAVIVideoFrame(SSurface* source_surface, int width, int height, bool8 sixteen_bit);
 
 static void S9xWinScanJoypads ();
 
@@ -332,8 +364,8 @@ bool8 SetupSound (long rate, bool8 sixteen_bit, bool8 stereo)
         if (DirectX.SetSoundMode ())
         {
             SoundBuffer = new uint8 [_blocksize * _buffernos + 1024 * 64];
+            ZeroMemory (SoundBuffer, _blocksize * _buffernos + 1024 * 64);
             SoundBuffer += 32 * 1024;
-            ZeroMemory (SoundBuffer, _blocksize * _buffernos);
             S9xSetPlaybackRate (so.playback_rate);
         }
         _samplecount = so.sixteen_bit ? _blocksize / 2 : _blocksize;
@@ -419,8 +451,8 @@ bool8 RealSetupSound (long rate, bool8 sixteen_bit, bool8 stereo)
     if (DirectX.SetSoundMode ())
     {
         SoundBuffer = new uint8 [_blocksize * _buffernos + 1024 * 64];
+        ZeroMemory (SoundBuffer, _blocksize * _buffernos + 1024 * 64);
         SoundBuffer += 32 * 1024;
-        ZeroMemory (SoundBuffer, _blocksize * _buffernos);
         S9xSetPlaybackRate (so.playback_rate);
     }
 
@@ -575,7 +607,7 @@ void ProcessSound (void)
             return;
 
         if( mute || Settings.ForcedPause || 
-            Settings.Paused || Settings.StopEmulation)
+            Settings.Paused || Settings.StopEmulation || GUI.AVIOut)
         {
             if (so.sixteen_bit)
             {
@@ -793,27 +825,16 @@ void S9xSyncSpeed( void)
     else
 #endif
 
-    if (!Settings.TurboMode && Settings.SkipFrames == AUTO_FRAMERATE)
+    if (!Settings.TurboMode && Settings.SkipFrames == AUTO_FRAMERATE &&
+		!GUI.AVIOut)
     {
-        LONG prev;
-        BOOL success;
-
-        if ((success = ReleaseSemaphore (GUI.FrameTimerSemaphore, 1, &prev)) &&
-            prev == 0)
+        if (!do_frame_adjust)
         {
-            // Mop up the ReleaseSemaphore above...
-            WaitForSingleObject (GUI.FrameTimerSemaphore, 0);
-            // ...and then wait for the real sync signal
-            WaitForSingleObject (GUI.FrameTimerSemaphore, 100);
             IPPU.RenderThisFrame = TRUE;
             IPPU.SkippedFrames = 0;
         }
         else
         {
-            if (success)
-                WaitForSingleObject (GUI.FrameTimerSemaphore, 0);
-            WaitForSingleObject (GUI.FrameTimerSemaphore, 0);
-
 	    if (IPPU.SkippedFrames < Settings.AutoMaxSkipFrames)
 	    {
 		IPPU.SkippedFrames++;
@@ -828,8 +849,12 @@ void S9xSyncSpeed( void)
     }
     else
     {
-	if (++IPPU.FrameSkip >= (Settings.TurboMode ? Settings.TurboSkipFrames
-                                                    : Settings.SkipFrames))
+	uint32 SkipFrames;
+	if(Settings.TurboMode)
+		SkipFrames = Settings.TurboSkipFrames;
+	else
+		SkipFrames = (Settings.SkipFrames == AUTO_FRAMERATE) ? 0 : Settings.SkipFrames;
+	if (++IPPU.FrameSkip >= SkipFrames)
 	{
 	    IPPU.FrameSkip = 0;
 	    IPPU.SkippedFrames = 0;
@@ -1297,7 +1322,8 @@ bool8 S9xDeinitUpdate (int Width, int Height, bool8 sixteen_bit)
     Src.Pitch = GFX.Pitch;
     Src.Surface = GFX.Screen;
 
-
+	// avi writing
+	DoAVIVideoFrame(&Src, Width, Height, sixteen_bit);
 
 	if(Settings.SixteenBit)
 	{
@@ -1320,6 +1346,8 @@ bool8 S9xDeinitUpdate (int Width, int Height, bool8 sixteen_bit)
 		Height = Src.Height;
 	}
 
+	GUI.ScreenCleared = true;
+
     SelectRenderMethod ();
 
     if (!VOODOO_MODE && !OPENGL_MODE)
@@ -1410,11 +1438,17 @@ bool8 S9xDeinitUpdate (int Width, int Height, bool8 sixteen_bit)
             
             if (GUI.Stretch)
             {	
-				p.x = p.y = 0;
+			/*	p.x = p.y = 0;
 				
                 ClientToScreen (GUI.hWnd, &p);
 				GetClientRect (GUI.hWnd, &dstRect);
 				OffsetRect(&dstRect, p.x, p.y);
+				*/
+				p.x = p.y = 0;
+				ClientToScreen (GUI.hWnd, &p);
+				GetClientRect (GUI.hWnd, &dstRect);
+				dstRect.bottom = int(double(dstRect.bottom) * double(239.0 / 240.0));
+				OffsetRect(&dstRect, p.x, p.y);
             }
             else
             {
@@ -1553,6 +1587,9 @@ void InitSnes9X( void)
     ScreenBuffer = ScreenBuf1 + EXT_OFFSET;
     memset (ScreenBuf1, 0, EXT_PITCH * EXT_HEIGHT);
     memset (ScreenBuf2, 0, EXT_PITCH * EXT_HEIGHT);
+    memset (SubScreenBuffer, 0, EXT_PITCH * EXT_HEIGHT);
+    memset (ZBuffer, 0, EXT_WIDTH * EXT_HEIGHT);
+    memset (SubZBuffer, 0, EXT_WIDTH * EXT_HEIGHT);
 
     GFX.Pitch = EXT_PITCH;
     GFX.RealPitch = EXT_PITCH;
@@ -1578,6 +1615,8 @@ void InitSnes9X( void)
     Settings.GlideEnable = FALSE;
 #endif
 
+	S9xMovieInit ();
+
     for (int C = 0; C != 16; C ++)
         Joystick[C].Attached = joyGetDevCaps (JOYSTICKID1+C, &Joystick[C].Caps,
                                               sizeof( JOYCAPS)) == JOYERR_NOERROR;
@@ -1608,6 +1647,7 @@ void DeinitS9x()
 	if(GUI.Accelerators)
 		DestroyAcceleratorTable(GUI.Accelerators);// = LoadAccelerators (hInstance, MAKEINTRESOURCE (IDR_SNES9X_ACCELERATORS));
 
+	S9xMovieStop (TRUE);
 }
 }
 int ffs (uint32 mask)
@@ -1825,7 +1865,7 @@ void S9xSetWinPixelFormat ()
 	}
 	maxdiff = newmaxdiff;
     }
-    delete color_diff;
+    delete [] color_diff;
 }
 
 void Convert8To24 (SSurface *src, SSurface *dst, RECT *srect)
@@ -2287,3 +2327,177 @@ void ResampleTo8000HzS8(uint8* input, uint8*output,int output_samples)
 }
 
 #endif
+
+void DoAVIOpen(const char* filename)
+{
+	// close current instance
+	if(GUI.AVIOut)
+	{
+		AVIClose(&GUI.AVIOut);
+		GUI.AVIOut = NULL;
+	}
+
+	// create new writer
+	AVICreate(&GUI.AVIOut);
+
+	int framerate = (Settings.PAL) ? 50 : 60;
+	int frameskip = Settings.SkipFrames;
+	if(frameskip == 0 || frameskip == AUTO_FRAMERATE)
+		frameskip = 1;
+
+	AVISetFramerate(framerate, frameskip, GUI.AVIOut);
+
+	BITMAPINFOHEADER bi;
+	memset(&bi, 0, sizeof(bi));
+	bi.biSize = 0x28;    
+	bi.biPlanes = 1;
+	bi.biBitCount = 24;
+	bi.biWidth = IPPU.RenderedScreenWidth;
+	bi.biHeight = IPPU.RenderedScreenHeight;
+	bi.biSizeImage = 3*bi.biWidth*bi.biHeight;
+
+	AVISetVideoFormat(&bi, GUI.AVIOut);
+
+	if(!Settings.Mute && SoundBuffer)
+	{
+		AVISetSoundFormat(&DirectX.wfx, GUI.AVIOut);
+	}
+
+	if(!AVIBegin(filename, GUI.AVIOut))
+	{
+		AVIClose(&GUI.AVIOut);
+		GUI.AVIOut = NULL;
+		return;
+	}
+
+	// init buffers
+	avi_sound_samples_per_update = (DirectX.wfx.nSamplesPerSec * frameskip) / framerate;
+	avi_sound_bytes_per_sample = DirectX.wfx.nBlockAlign;
+
+	avi_buffer = new uint8[3*IPPU.RenderedScreenWidth*IPPU.RenderedScreenHeight];
+	avi_sound_buffer = new uint8[avi_sound_samples_per_update * avi_sound_bytes_per_sample];
+
+	avi_width = IPPU.RenderedScreenWidth;
+	avi_height = IPPU.RenderedScreenHeight;
+	avi_skip_frames = Settings.SkipFrames;
+}
+
+void DoAVIClose(int reason)
+{
+	if(!GUI.AVIOut)
+	{
+		return;
+	}
+
+	AVIClose(&GUI.AVIOut);
+	GUI.AVIOut = NULL;
+
+	delete [] avi_buffer;
+	delete [] avi_sound_buffer;
+
+	avi_buffer = NULL;
+	avi_sound_buffer = NULL;
+
+	switch(reason)
+	{
+	case 1:
+		// emu settings changed
+		S9xMessage(S9X_INFO, S9X_AVI_INFO, AVI_CONFIGURATION_CHANGED);
+		break;
+	default:
+		// print nothing
+		break;
+	}
+}
+
+static void DoAVIVideoFrame(SSurface* source_surface, int Width, int Height, bool8 sixteen_bit)
+{
+	if(!GUI.AVIOut)
+	{
+		return;
+	}
+
+	// check configuration
+	const WAVEFORMATEX* pwfex = NULL;
+	if(avi_width != Width ||
+		avi_height != Height ||
+		avi_skip_frames != Settings.SkipFrames ||
+		(AVIGetSoundFormat(GUI.AVIOut, &pwfex) && memcmp(pwfex, &DirectX.wfx, sizeof(WAVEFORMATEX))))
+	{
+		DoAVIClose(1);
+		return;
+	}
+	
+	// convert to bitdepth 24
+	SSurface avi_dest_surface;
+	RECT full_rect;
+	avi_dest_surface.Surface = avi_buffer;
+	avi_dest_surface.Pitch = Width * 3;
+	avi_dest_surface.Width = Width;
+	avi_dest_surface.Height = Height;
+	full_rect.top = 0;
+	full_rect.left = 0;
+	full_rect.bottom = Height;
+	full_rect.right = Width;
+	if(sixteen_bit)
+	{
+		Convert16To24Packed(source_surface, &avi_dest_surface, &full_rect);
+	}
+	else
+	{
+		Convert8To24Packed(source_surface, &avi_dest_surface, &full_rect);
+	}
+
+	// flip the image vertically
+	const int pitch = 3*Width;
+	int y;
+	for(y=0; y<Height>>1; ++y)
+	{
+		uint8* lo_8 = avi_buffer+y*pitch;
+		uint8* hi_8 = avi_buffer+(Height-1-y)*pitch;
+		uint32* lo_32=(uint32*)lo_8;
+		uint32* hi_32=(uint32*)hi_8;
+
+		int q;
+		{
+			register uint32 a, b;
+			for(q=pitch>>4; q>0; --q)
+			{
+				a=*lo_32;  b=*hi_32;  *lo_32=b;  *hi_32=a;  ++lo_32;  ++hi_32;
+				a=*lo_32;  b=*hi_32;  *lo_32=b;  *hi_32=a;  ++lo_32;  ++hi_32;
+				a=*lo_32;  b=*hi_32;  *lo_32=b;  *hi_32=a;  ++lo_32;  ++hi_32;
+				a=*lo_32;  b=*hi_32;  *lo_32=b;  *hi_32=a;  ++lo_32;  ++hi_32;
+			}
+		}
+
+		{
+			register uint8 c, d;
+			for(q=(pitch&0x0f); q>0; --q)
+			{
+				c=*lo_8;  d=*hi_8;  *lo_8=d;  *hi_8=c;
+			}
+		}
+	}
+
+	// write to AVI
+	AVIAddVideoFrame(avi_buffer, GUI.AVIOut);
+
+	// generate sound
+	if(pwfex)
+	{
+		const int _maxsamplecount = SOUND_BUFFER_SIZE / avi_sound_bytes_per_sample;
+		const int stereo_multiplier = (Settings.Stereo) ? 2 : 1;
+		int samples_written = 0;
+		int sample_count = avi_sound_samples_per_update;
+		while(sample_count > _maxsamplecount)
+		{
+			S9xMixSamples(avi_sound_buffer+(samples_written*avi_sound_bytes_per_sample), _maxsamplecount*stereo_multiplier);
+			sample_count -= _maxsamplecount;
+			samples_written += _maxsamplecount;
+		}
+		S9xMixSamples(avi_sound_buffer+(samples_written*avi_sound_bytes_per_sample), sample_count*stereo_multiplier);
+
+		AVIAddSoundSamples(avi_sound_buffer, avi_sound_samples_per_update, GUI.AVIOut);
+	}
+}
+
